import math
from pathlib import Path
from typing import Generator, Optional, Tuple

import numpy as np
import numpy.typing as npt
from plyfile import PlyData, PlyElement
from pyproj import Transformer

from py3dtiles.typing import (
    MetadataReaderType,
    OffsetScaleType,
    PortionItemType,
    PortionsType,
)


def get_metadata(path: Path, fraction: int = 100) -> MetadataReaderType:
    """Get metadata in case of a input ply file."""
    ply_point_cloud = PlyData.read(path)
    if "vertex" not in [e.name for e in ply_point_cloud.elements]:
        raise KeyError(
            "The ply data does not contain any 'vertex' item. Are you sure the file is valid?"
        )
    ply_vertices = ply_point_cloud["vertex"]
    point_count = ply_vertices.count * fraction // 100
    ply_features = [ply_prop.name for ply_prop in ply_vertices.properties]
    if any(coord not in ply_features for coord in ("x", "y", "z")):
        raise KeyError(
            "At least one of the basic coordinate feature (x, y, z) is missing in the input file."
        )

    data = np.array(
        [ply_vertices["x"], ply_vertices["y"], ply_vertices["z"]]
    ).transpose()
    aabb = np.array((np.min(data, axis=0), np.max(data, axis=0)))

    pointcloud_file_portions: PortionsType = [(str(path), (0, point_count))]

    return {
        "portions": pointcloud_file_portions,
        "aabb": aabb,
        "crs_in": None,
        "point_count": point_count,
        "avg_min": aabb[0],
    }


def run(
    filename: str,
    offset_scale: OffsetScaleType,
    portion: PortionItemType,
    transformer: Optional[Transformer],
    color_scale: Optional[float],
    write_intensity: bool,
) -> Generator[
    Tuple[
        npt.NDArray[np.float32],
        npt.NDArray[np.uint8],
        npt.NDArray[np.uint8],
        npt.NDArray[np.uint8],
    ],
    None,
    None,
]:
    """
    Reads points from a ply file.
    """
    ply_point_cloud = PlyData.read(filename)
    ply_vertices = ply_point_cloud["vertex"]

    point_count = portion[1] - portion[0]
    step = min(point_count, max(point_count // 10, 100_000))
    indices = list(range(math.ceil(point_count / step)))

    for index in indices:
        start_offset = portion[0] + index * step
        num = min(step, portion[1] - start_offset)

        x = ply_vertices["x"][start_offset : (start_offset + num)]
        y = ply_vertices["y"][start_offset : (start_offset + num)]
        z = ply_vertices["z"][start_offset : (start_offset + num)]
        if transformer:
            x, y, z = transformer.transform(x, y, z)

        x = (x + offset_scale[0][0]) * offset_scale[1][0]
        y = (y + offset_scale[0][1]) * offset_scale[1][1]
        z = (z + offset_scale[0][2]) * offset_scale[1][2]

        coords = np.vstack((x, y, z)).transpose()

        if offset_scale[2] is not None:
            # Apply transformation matrix (because the tile's transform will contain
            # the inverse of this matrix)
            coords = np.dot(coords, offset_scale[2])

        coords = np.ascontiguousarray(coords.astype(np.float32))

        # NOTE this code assume all the colors have the same type
        # I think it's a reasonable assumption to make at this point but it's
        # not mandated by the spec!
        if "red" in ply_vertices:
            # val_dtype is of the form: i<nbytes>, u<nbytes>, float<nbytes>
            # see https://github.com/dranjan/python-plyfile/blob/d1f73004ed0a296fc8b9c1fad8139b5d90410639/plyfile.py#L32
            signed = ply_vertices.ply_property("red").val_dtype[0:1] == "i"
            nbytes = int(ply_vertices.ply_property("red").val_dtype[1:])
            # factor is how much we need to divide the data
            # nbytes - 1 is the number of "extra" bytes compared to the 1 byte target
            # but we remove one extra bit if it is signed
            # max, because if the value is *shorter* we don't need to do anything)
            factor = max(1, 2 ** (8 * (nbytes - 1) - (1 if signed else 0)))
            red = ply_vertices["red"][start_offset : (start_offset + num)] / factor
            green = ply_vertices["green"][start_offset : (start_offset + num)] / factor
            blue = ply_vertices["blue"][start_offset : (start_offset + num)] / factor
        else:
            red = green = blue = np.zeros(num)

        raw_colors = np.vstack((red, green, blue)).transpose()

        if color_scale is not None:
            raw_colors = np.clip(raw_colors * color_scale, 0, 255)

        colors = raw_colors.astype(np.uint8)

        if "classification" in ply_vertices:
            classification = np.array(
                ply_vertices["classification"].reshape(-1, 1), dtype=np.uint8
            )
        else:
            classification = np.zeros((coords.shape[0], 1), dtype=np.uint8)

        if "intensity" in ply_vertices and write_intensity:
            if ply_vertices["intensity"].dtype != np.uint8:
                print(
                    "Warning: At the moment, only intensity in uint8 format is supported for ply files"
                )
                intensity = np.zeros((coords.shape[0], 1), dtype=np.uint8)
            else:
                intensity = np.array(
                    ply_vertices["intensity"].reshape(-1, 1), dtype=np.uint8
                )
        else:
            intensity = np.zeros((coords.shape[0], 1), dtype=np.uint8)

        yield coords, colors, classification, intensity


def create_plydata_with_renamed_property(
    plydata: PlyData, old_property_name: str, new_property_name: str
) -> PlyData:
    """Create a new plyfile.PlyData object on the model of the one provided as input, with a modified
    feature name.

    This function may be useful for handling classification feature name in .ply files, knowing
    that this feature is not formalized.

    """
    ply_data = plydata["vertex"].data
    copied_data = ply_data.copy()
    ptype = np.dtype(
        [
            (new_property_name, t[1]) if t[0] == old_property_name else t
            for t in ply_data.dtype.descr
        ]
    )
    pelement = PlyElement.describe(data=copied_data.astype(ptype), name="vertex")
    return PlyData(elements=[pelement])



"""
This package contains all the common files for tilers, especially abstract base
classes a tiler must derive.

# How to create a tiler?

You should start by deriving the Tiler class.
"""
from .shared_metadata import SharedMetadata
from .tiler import Tiler
from .tiler_worker import TilerWorker

__all__ = ["SharedMetadata", "Tiler", "TilerWorker"]



import argparse
import os
import pickle
import shutil
import sys
import tempfile
import time
import traceback
from multiprocessing import Process, cpu_count
from pathlib import Path
from time import sleep
from typing import Any, Dict, List, Optional, Set, Union

import psutil
import zmq
from pyproj import CRS

from py3dtiles.constants import EXIT_CODES
from py3dtiles.exceptions import (
    Py3dtilesException,
    SrsInMissingException,
    TilerException,
    WorkerException,
)
from py3dtiles.tilers.base_tiler import Tiler
from py3dtiles.tilers.base_tiler.message_type import ManagerMessage, WorkerMessageType
from py3dtiles.tilers.base_tiler.tiler_worker import TilerWorker
from py3dtiles.tilers.point.point_tiler import PointTiler
from py3dtiles.utils import mkdir_or_raise, str_to_CRS

TOTAL_MEMORY_MB = int(psutil.virtual_memory().total / (1024 * 1024))
DEFAULT_CACHE_SIZE = int(TOTAL_MEMORY_MB / 10)
CPU_COUNT = cpu_count()

# IPC protocol is not supported on Windows
if os.name == "nt":
    URI = "tcp://127.0.0.1:0"
else:
    # Generate a unique name for this socket
    tmpdir = tempfile.TemporaryDirectory()
    URI = f"ipc://{tmpdir.name}/py3dtiles.sock"


META_TILER_NAME = b"meta"


def _worker_target(
    worker_tilers: Dict[bytes, TilerWorker[Any]],
    verbosity: int,
    uri: bytes,
) -> None:
    return _WorkerDispatcher(
        worker_tilers,
        verbosity,
        uri,
    ).run()


class _WorkerDispatcher:
    """
    This class waits from jobs commands from the Zmq socket.
    """

    skt: zmq.Socket[bytes]

    def __init__(
        self,
        worker_tilers: Dict[bytes, TilerWorker[Any]],
        verbosity: int,
        uri: bytes,
    ) -> None:
        self.worker_tilers = worker_tilers
        self.verbosity = verbosity
        self.uri = uri

        # Socket to receive messages on
        self.context = zmq.Context()

    def run(self) -> None:
        self.skt = self.context.socket(zmq.DEALER)

        self.skt.connect(self.uri)  # type: ignore [arg-type]

        startup_time = time.time()
        idle_time = 0.0

        # notify we're ready
        self.skt.send_multipart([WorkerMessageType.REGISTER.value])

        while True:
            try:
                before = time.time() - startup_time
                self.skt.poll()
                after = time.time() - startup_time

                idle_time += after - before

                message = self.skt.recv_multipart()
                tiler_name = message[1]
                command = message[2]
                content = message[3:]

                delta = time.time() - pickle.loads(message[0])
                if delta > 0.01 and self.verbosity >= 1:
                    print(
                        f"{os.getpid()} / {round(after, 2)} : Delta time: {round(delta, 3)}"
                    )

                if command == ManagerMessage.SHUTDOWN.value:
                    break  # ack
                else:
                    self.worker_tilers[tiler_name].execute(self.skt, command, content)

                # notify we're idle
                self.skt.send_multipart([WorkerMessageType.IDLE.value])
            except Exception as e:
                traceback.print_exc()
                error_message = f"{e.__class__.__module__}.{e.__class__.__name__}: {e}"
                self.skt.send_multipart(
                    [WorkerMessageType.ERROR.value, error_message.encode()]
                )
                # we still print it for stacktraces

        if self.verbosity >= 1:
            print(
                "total: {} sec, idle: {}".format(
                    round(time.time() - startup_time, 1), round(idle_time, 1)
                )
            )

        self.skt.send_multipart([WorkerMessageType.HALTED.value])


# Manager
class _ZmqManager:
    """
    This class sends messages to the workers.
    We can also request general status.
    """

    def __init__(
        self,
        number_of_jobs: int,
        worker_tilers: Dict[bytes, TilerWorker[Any]],
        verbosity: int,
    ) -> None:
        """
        For the process_args argument, see the init method of Worker
        to get the list of needed parameters.
        """
        self.context = zmq.Context()

        self.number_of_jobs = number_of_jobs

        self.socket = self.context.socket(zmq.ROUTER)
        self.socket.bind(URI)
        # Useful only when TCP is used to get the URI with the opened port
        self.uri = self.socket.getsockopt(zmq.LAST_ENDPOINT)
        if not isinstance(self.uri, bytes):
            raise RuntimeError(
                "The uri returned by self.socket.getsockopt should be bytes."
            )

        self.processes = [
            Process(
                target=_worker_target,
                args=(worker_tilers, verbosity, self.uri),
            )
            for _ in range(number_of_jobs)
        ]
        for p in self.processes:
            p.start()

        self.activities = [p.pid for p in self.processes]
        self.clients: Set[bytes] = set()
        self.idle_clients: Set[bytes] = set()

        self.killing_processes = False
        self.number_processes_killed = 0
        self.time_waiting_an_idle_process = 0.0

    def all_clients_registered(self) -> bool:
        return len(self.clients) == self.number_of_jobs

    def send_to_process(self, message: List[bytes]) -> None:
        if not self.idle_clients:
            raise ValueError("idle_clients is empty")
        self.socket.send_multipart(
            [self.idle_clients.pop(), pickle.dumps(time.time())] + message
        )

    def send_to_all_processes(self, message: List[bytes]) -> None:
        if len(self.clients) == 0:
            raise ValueError("No registered clients")
        for client in self.clients:
            self.socket.send_multipart([client, pickle.dumps(time.time())] + message)

    def send_to_all_idle_processes(self, message: List[bytes]) -> None:
        if not self.idle_clients:
            raise ValueError("idle_clients is empty")
        for client in self.idle_clients:
            self.socket.send_multipart([client, pickle.dumps(time.time())] + message)
        self.idle_clients.clear()

    def can_queue_more_jobs(self) -> bool:
        return len(self.idle_clients) != 0

    def register_client(self, client_id: bytes) -> None:
        if client_id in self.clients:
            print(f"Warning: {client_id!r} already registered")
        else:
            self.clients.add(client_id)
        self.add_idle_client(client_id)

    def add_idle_client(self, client_id: bytes) -> None:
        if client_id in self.idle_clients:
            raise ValueError(f"The client id {client_id!r} is already in idle_clients")
        self.idle_clients.add(client_id)

    def are_all_processes_idle(self) -> bool:
        return len(self.idle_clients) == self.number_of_jobs

    def are_all_processes_killed(self) -> bool:
        return self.number_processes_killed == self.number_of_jobs

    def shutdown_all_processes(self) -> None:
        self.send_to_all_processes([META_TILER_NAME, ManagerMessage.SHUTDOWN.value])
        self.killing_processes = True

    def join_all_processes(self) -> None:
        for p in self.processes:
            p.join()


def convert(
    files: Union[List[Union[str, Path]], str, Path],
    outfolder: Union[str, Path] = "./3dtiles",
    overwrite: bool = False,
    jobs: int = CPU_COUNT,
    cache_size: int = DEFAULT_CACHE_SIZE,
    crs_out: Optional[CRS] = None,
    crs_in: Optional[CRS] = None,
    force_crs_in: bool = False,
    benchmark: Optional[str] = None,
    rgb: bool = True,
    classification: bool = True,
    intensity: bool = True,
    color_scale: Optional[float] = None,
    verbose: int = False,
) -> None:
    """
    Convert the input dataset into 3dtiles. For the argument list and their effects, please see :py:class:`.Converter`.

    :param files: Filenames to process. The file must use the .las, .laz, .xyz or .ply format.
    :param outfolder: The folder where the resulting tileset will be written.
    :param overwrite: Overwrite the ouput folder if it already exists.
    :param jobs: The number of parallel jobs to start. Default to the number of cpu.
    :param cache_size: Cache size in MB. Default to available memory / 10.
    :param crs_out: CRS to convert the output with
    :param crs_in: Set a default input CRS
    :param force_crs_in: Force every input CRS to be `crs_in`, even if not null
    :param benchmark: Print summary at the end of the process
    :param rgb: Export rgb attributes.
    :param classification: Export classification attribute.
    :param intensity: Export intensity attributes. This support is currently limited to unsigned 8 bits integer for ply files, and to integers for xyz files.
    :param color_scale: Scale the color with the specified amount. Useful to lighten or darken black pointclouds with only intensity.

    :raises SrsInMissingException: if py3dtiles couldn't find srs informations in input files and srs_in is not specified
    :raises SrsInMixinException: if the input files have different CRS

    """
    converter = _Convert(
        files,
        outfolder=outfolder,
        overwrite=overwrite,
        jobs=jobs,
        cache_size=cache_size,
        crs_out=crs_out,
        crs_in=crs_in,
        force_crs_in=force_crs_in,
        benchmark=benchmark,
        rgb=rgb,
        classification=classification,
        intensity=intensity,
        color_scale=color_scale,
        verbose=verbose,
    )
    return converter.convert()


class _Convert:
    def __init__(
        self,
        files: Union[List[Union[str, Path]], str, Path],
        outfolder: Union[str, Path] = "./3dtiles",
        overwrite: bool = False,
        jobs: int = CPU_COUNT,
        cache_size: int = DEFAULT_CACHE_SIZE,
        crs_out: Optional[CRS] = None,
        crs_in: Optional[CRS] = None,
        force_crs_in: bool = False,
        benchmark: Optional[str] = None,
        rgb: bool = True,
        classification: bool = True,
        intensity: bool = True,
        color_scale: Optional[float] = None,
        verbose: int = False,
    ) -> None:
        """
        :param files: Filenames to process. The file must use the .las, .laz, .xyz or .ply format.
        :param outfolder: The folder where the resulting tileset will be written.
        :param overwrite: Overwrite the ouput folder if it already exists.
        :param jobs: The number of parallel jobs to start. Default to the number of cpu.
        :param cache_size: Cache size in MB. Default to available memory / 10.
        :param crs_out: CRS to convert the output with
        :param crs_in: Set a default input CRS
        :param force_crs_in: Force every input CRS to be `crs_in`, even if not null
        :param benchmark: Print summary at the end of the process
        :param rgb: Export rgb attributes.
        :param classification: Export classification attribute.
        :param intensity: Export intensity attribute.
        :param color_scale: Scale the color with the specified amount. Useful to lighten or darken black pointclouds with only intensity.

        :raises SrsInMissingException: if py3dtiles couldn't find srs informations in input files and srs_in is not specified
        :raises SrsInMixinException: if the input files have different CRS

        """
        # create folder
        self.out_folder = Path(outfolder)
        mkdir_or_raise(self.out_folder, overwrite=overwrite)

        self.tilers = [
            PointTiler(
                self.out_folder,
                files,
                crs_in,
                force_crs_in,
                rgb,
                classification,
                intensity,
                color_scale,
                cache_size,
                verbose,
            )
        ]

        self.jobs = jobs

        self.verbose = verbose
        self.benchmark = benchmark

        self.working_dir = self.out_folder / "tmp"
        self.working_dir.mkdir(parents=True)

        worker_tilers: Dict[bytes, TilerWorker[Any]] = {}
        for tiler in self.tilers:
            if tiler.name in worker_tilers:
                raise TilerException("There are tilers with the same attribute name.")

            try:
                tiler.initialization(
                    crs_out, self.working_dir / str(tiler.name), self.jobs
                )
            except Py3dtilesException as e:
                shutil.rmtree(self.out_folder)
                raise e

            worker_tilers[tiler.name] = tiler.get_worker()

        if self.verbose >= 1:
            for tiler in self.tilers:
                tiler.print_summary()

        self.zmq_manager = _ZmqManager(
            self.jobs,
            worker_tilers,
            self.verbose,
        )

    def convert(self) -> None:
        """convert

        Convert pointclouds (xyz, las or laz) to 3dtiles tileset containing pnts node
        """
        startup: float = time.time()

        try:
            for tiler in self.tilers:
                while True:
                    now = time.time() - startup

                    at_least_one_job_ended = False
                    if (
                        not self.zmq_manager.can_queue_more_jobs()
                        or self.zmq_manager.socket.poll(timeout=0, flags=zmq.POLLIN)
                    ):
                        at_least_one_job_ended = self.process_message(tiler)

                    # we wait for all processes/threads to register
                    # if we don't there are tricky cases where an exception fires in a worker before all the workers registered, which means that not all workers will receive the shutdown signal
                    if not self.zmq_manager.all_clients_registered():
                        sleep(0.1)
                        continue

                    if self.zmq_manager.can_queue_more_jobs():
                        for command, data in tiler.get_tasks(startup):
                            self.zmq_manager.send_to_process(
                                [PointTiler.name, command] + data
                            )
                            if not self.zmq_manager.can_queue_more_jobs():
                                break

                    # if at this point we have no work in progress => we're done
                    if self.zmq_manager.are_all_processes_idle():
                        break

                    if at_least_one_job_ended:
                        tiler.print_debug(
                            now, self.jobs, len(self.zmq_manager.idle_clients)
                        )

                    tiler.memory_control()

                tiler.validate_binary_data()

                if self.verbose >= 1:
                    print("Writing 3dtiles")

                tiler.write_tileset()
                shutil.rmtree(self.working_dir / str(tiler.name), ignore_errors=True)

                if self.verbose >= 1:
                    print(f"Tiler {tiler.name!r} done")

                if self.benchmark:
                    tiler.benchmark(self.benchmark, startup)

        finally:
            self.zmq_manager.shutdown_all_processes()
            self.zmq_manager.join_all_processes()

            if self.verbose >= 1:
                print(
                    "destroy", round(self.zmq_manager.time_waiting_an_idle_process, 2)
                )

            self.zmq_manager.context.destroy()

    def process_message(self, tiler: Tiler[Any, Any]) -> bool:
        at_least_one_job_ended = False

        # Blocking read but it's fine because either all our child processes are busy
        # or we know that there's something to read (zmq.POLLIN)
        start = time.time()
        message = self.zmq_manager.socket.recv_multipart()

        client_id = message[0]
        return_type = message[1]
        content = message[2:]

        if return_type == WorkerMessageType.REGISTER.value:
            self.zmq_manager.register_client(client_id)
        elif return_type == WorkerMessageType.IDLE.value:
            self.zmq_manager.add_idle_client(client_id)

            if not self.zmq_manager.can_queue_more_jobs():
                self.zmq_manager.time_waiting_an_idle_process += time.time() - start

        elif return_type == WorkerMessageType.HALTED.value:
            self.zmq_manager.number_processes_killed += 1

        elif return_type == WorkerMessageType.ERROR.value:
            raise WorkerException(
                f"An exception occurred in a worker: {content[0].decode()}"
            )

        else:
            at_least_one_job_ended = tiler.process_message(return_type, content)

        return at_least_one_job_ended


def _init_parser(
    subparser: "argparse._SubParsersAction[Any]",
) -> argparse.ArgumentParser:
    parser: argparse.ArgumentParser = subparser.add_parser(
        "convert",
        help="Convert input 3D data to a 3dtiles tileset.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "files",
        nargs="+",
        help="Filenames to process. The file must use the .las, .laz (lastools must be installed), .xyz or .ply format.",
    )
    parser.add_argument(
        "--out",
        type=str,
        help="The folder where the resulting tileset will be written.",
        default="./3dtiles",
    )
    parser.add_argument(
        "--overwrite",
        help="Delete and recreate the ouput folder if it already exists. WARNING: be careful, there will be no confirmation!",
        action="store_true",
    )
    parser.add_argument(
        "--jobs",
        help="The number of parallel jobs to start. Default to the number of cpu.",
        default=cpu_count(),
        type=int,
    )
    parser.add_argument(
        "--cache_size",
        help="Cache size in MB. Default to available memory / 10.",
        default=int(TOTAL_MEMORY_MB / 10),
        type=int,
    )
    parser.add_argument(
        "--srs_out",
        help="SRS to convert the output with (numeric part of the EPSG code)",
        type=str,
    )
    parser.add_argument(
        "--srs_in", help="Override input SRS (numeric part of the EPSG code)", type=str
    )
    parser.add_argument(
        "--benchmark", help="Print summary at the end of the process", type=str
    )
    parser.add_argument(
        "--no-rgb", help="Don't export rgb attributes", action="store_true"
    )
    parser.add_argument(
        "--classification", help="Export classification attributes", action="store_true"
    )
    parser.add_argument(
        "--intensity",
        help="Export intensity attributes. This support is currently limited to unsigned 8 bits integer for ply files, and to integers for xyz files.",
        action="store_true",
    )
    parser.add_argument("--color_scale", help="Force color scale", type=float)
    parser.add_argument(
        "--force-srs-in",
        help="Force the input srs even if the srs in the input files are different. CAUTION, use only if you know what you are doing.",
        action="store_true",
    )

    return parser


def _main(args: argparse.Namespace) -> None:
    try:
        return convert(
            args.files,
            outfolder=args.out,
            overwrite=args.overwrite,
            jobs=args.jobs,
            cache_size=args.cache_size,
            crs_out=str_to_CRS(args.srs_out),
            crs_in=str_to_CRS(args.srs_in),
            force_crs_in=args.force_srs_in,
            benchmark=args.benchmark,
            rgb=not args.no_rgb,
            classification=args.classification,
            intensity=args.intensity,
            color_scale=args.color_scale,
            verbose=args.verbose,
        )
    except SrsInMissingException:
        print(
            "No SRS information in input files, you should specify it with --srs_in",
            file=sys.stderr,
        )
        sys.exit(EXIT_CODES.MISSING_SRS_IN_FILE.value)



from pathlib import Path

import numpy as np
from numpy.testing import assert_array_equal

from py3dtiles.tilers.point.node.node import split_tileset
from py3dtiles.tileset.bounding_volume_box import BoundingVolumeBox
from py3dtiles.tileset.tile import Tile
from py3dtiles.tileset.tileset import TileSet


def test_split_tileset(tmp_dir: Path) -> None:
    tile = Tile(geometric_error=1234)
    tile.bounding_volume = BoundingVolumeBox.from_points(
        [np.array([1, 2, 4]), np.array([4, 8, 0])]
    )
    tile.set_refine_mode("REPLACE")

    split_tileset(tile, "split", tmp_dir)
    assert tile.children == []
    assert tile.content_uri == Path("tileset.split.json")

    tileset_path = tmp_dir / "tileset.split.json"
    assert tileset_path.exists()
    ts = TileSet.from_file(tileset_path)
    assert ts.root_tile.geometric_error == 1234
    # these 3 lines are *mainly* to make mypy happy
    assert isinstance(ts.root_tile.bounding_volume, BoundingVolumeBox)
    assert ts.root_tile.bounding_volume._box is not None
    assert tile.bounding_volume._box is not None

    assert_array_equal(ts.root_tile.bounding_volume._box, tile.bounding_volume._box)
    # sure about this one?
    assert ts.root_tile.get_refine_mode() == "ADD"



from __future__ import annotations

import math
import pickle

import lz4.frame as gzip
import numpy as np
import numpy.typing as npt

from py3dtiles.tilers.point.node.node import Node
from py3dtiles.utils import split_aabb


class NodeCatalog:
    """NodeCatalog is a store of Node objects.py3dtiles

    Using a NodeCatalog allows to only store a children names
    in nodes, instead of storing a full recursive structure.
    """

    def __init__(
        self,
        nodes: bytes,
        name: bytes,
        root_aabb: npt.NDArray[np.float64],
        root_spacing: float,
    ) -> None:
        self.nodes: dict[bytes, Node] = {}
        self.root_aabb = root_aabb
        self.root_spacing = root_spacing
        self.node_bytes: dict[bytes, bytes] = {}
        self._load_from_store(name, nodes)

    def get_node(self, name: bytes) -> Node:
        """Returns the node mathing the given name"""
        if name not in self.nodes:
            spacing = self.root_spacing / math.pow(2, len(name))
            aabb = self.root_aabb
            for i in name:
                aabb = split_aabb(aabb, int(i))
            node = Node(name, aabb, spacing)
            self.nodes[name] = node
        else:
            node = self.nodes[name]
        return node

    def dump(self, name: bytes, max_depth: int) -> bytes:
        """Serialize the stored nodes to a bytes list"""
        node = self.nodes[name]
        if node.dirty:
            self.node_bytes[name] = node.save_to_bytes()

        if node.children is not None and max_depth > 0:
            for n in node.children:
                self.dump(n, max_depth - 1)

        return pickle.dumps(self.node_bytes)

    def _load_from_store(self, name: bytes, data: bytes) -> Node:
        if len(data) > 0:
            out = pickle.loads(gzip.decompress(data))
            for n in out:
                spacing = self.root_spacing / math.pow(2, len(n))
                aabb = self.root_aabb
                for i in n:
                    aabb = split_aabb(aabb, int(i))
                node = Node(n, aabb, spacing)
                node.load_from_bytes(out[n])
                self.node_bytes[n] = out[n]
                self.nodes[n] = node
        else:
            spacing = self.root_spacing / math.pow(2, len(name))
            aabb = self.root_aabb
            for i in name:
                aabb = split_aabb(aabb, int(i))
            node = Node(name, aabb, spacing)
            self.nodes[name] = node

        return self.nodes[name]



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar

import numpy as np
import numpy.typing as npt

from py3dtiles.typing import (
    BoundingVolumeBoxDictType,
    BoundingVolumeRegionDictType,
    BoundingVolumeSphereDictType,
)

from .root_property import RootProperty

if TYPE_CHECKING:
    from typing_extensions import Self

    from py3dtiles.tileset import Tile

_BoundingVolumeJsonDictT = TypeVar(
    "_BoundingVolumeJsonDictT",
    BoundingVolumeBoxDictType,
    BoundingVolumeRegionDictType,
    BoundingVolumeSphereDictType,
)


class BoundingVolume(
    RootProperty[_BoundingVolumeJsonDictT], Generic[_BoundingVolumeJsonDictT]
):
    """
    Abstract class used as interface for box, region and sphere
    """

    def __init__(self) -> None:
        super().__init__()

    @classmethod
    @abstractmethod
    def from_dict(cls, bounding_volume_dict: _BoundingVolumeJsonDictT) -> Self:
        ...

    @abstractmethod
    def get_center(self) -> npt.NDArray[np.float64]:
        ...

    @abstractmethod
    def translate(self, offset: npt.NDArray[np.float64]) -> None:
        ...

    @abstractmethod
    def transform(self, transform: npt.NDArray[np.float64]) -> None:
        ...

    @abstractmethod
    def add(self, other: BoundingVolume[Any]) -> None:
        ...

    @abstractmethod
    def sync_with_children(self, owner: Tile) -> None:
        ...

    @abstractmethod
    def to_dict(self) -> _BoundingVolumeJsonDictT:
        ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import numpy.typing as npt
from numba import njit  # type: ignore [attr-defined]
from numba.typed import List

from py3dtiles.exceptions import TilerException
from py3dtiles.utils import SubdivisionType, aabb_size_to_subdivision_type

from .distance import is_point_far_enough, xyz_to_key

if TYPE_CHECKING:
    from .node import Node


@njit(fastmath=True, cache=True)  # type: ignore [misc]
def _insert(
    cells_xyz: List[npt.NDArray[np.float32]],
    cells_rgb: List[npt.NDArray[np.uint8]],
    cells_classification: List[npt.NDArray[np.uint8]],
    cells_intensity: List[npt.NDArray[np.uint8]],
    aabmin: npt.NDArray[np.float32],
    inv_aabb_size: npt.NDArray[np.float32],
    cell_count: npt.NDArray[np.int32],
    xyz: npt.NDArray[np.float32],
    rgb: npt.NDArray[np.uint8],
    classification: npt.NDArray[np.uint8],
    intensity: npt.NDArray[np.uint8],
    spacing: float,
    shift: int,
    force: bool = False,
) -> tuple[
    npt.NDArray[np.float32],
    npt.NDArray[np.uint8],
    npt.NDArray[np.uint8],
    npt.NDArray[np.uint8],
    bool,
] | None:
    keys = xyz_to_key(xyz, cell_count, aabmin, inv_aabb_size, shift)

    if force:
        # allocate this one once and for all
        for k in np.unique(keys):
            idx = np.where(keys - k == 0)
            cells_xyz[k] = np.concatenate((cells_xyz[k], xyz[idx]))
            cells_rgb[k] = np.concatenate((cells_rgb[k], rgb[idx]))
            cells_classification[k] = np.concatenate(
                (cells_classification[k], classification[idx])
            )
            cells_intensity[k] = np.concatenate((cells_intensity[k], intensity[idx]))
        return None
    else:
        notinserted = np.full(len(xyz), False)
        needs_balance = False

        for i in range(len(xyz)):
            k = keys[i]
            if cells_xyz[k].shape[0] == 0 or is_point_far_enough(
                cells_xyz[k], xyz[i], spacing
            ):
                cells_xyz[k] = np.concatenate((cells_xyz[k], xyz[i].reshape(1, 3)))
                cells_rgb[k] = np.concatenate((cells_rgb[k], rgb[i].reshape(1, 3)))

                cells_classification[k] = np.concatenate(
                    (cells_classification[k], classification[i].reshape(1, 1))
                )
                cells_intensity[k] = np.concatenate(
                    (cells_intensity[k], intensity[i].reshape(1, 1))
                )
                if cell_count[0] < 8:
                    needs_balance = needs_balance or cells_xyz[k].shape[0] > 200000
            else:
                notinserted[i] = True

        return (
            xyz[notinserted],
            rgb[notinserted],
            classification[notinserted],
            intensity[notinserted],
            needs_balance,
        )


class Grid:
    """docstring for Grid"""

    __slots__ = (
        "cell_count",
        "cells_xyz",
        "cells_rgb",
        "cells_classification",
        "cells_intensity",
        "spacing",
    )

    def __init__(self, node: Node, initial_count: int = 3) -> None:
        self.cell_count = np.array(
            [initial_count, initial_count, initial_count], dtype=np.int32
        )
        self.spacing = node.spacing * node.spacing

        self.cells_xyz = List()
        self.cells_rgb = List()
        self.cells_classification = List()
        self.cells_intensity = List()
        for _ in range(self.max_key_value):
            self.cells_xyz.append(np.zeros((0, 3), dtype=np.float32))
            self.cells_rgb.append(np.zeros((0, 3), dtype=np.uint8))
            self.cells_classification.append(np.zeros((0, 1), dtype=np.uint8))
            self.cells_intensity.append(np.zeros((0, 1), dtype=np.uint8))

    def __getstate__(self) -> dict[str, Any]:
        return {
            "cell_count": self.cell_count,
            "spacing": self.spacing,
            "cells_xyz": list(self.cells_xyz),
            "cells_rgb": list(self.cells_rgb),
            "cells_classification": list(self.cells_classification),
            "cells_intensity": list(self.cells_intensity),
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.cell_count = state["cell_count"]
        self.spacing = state["spacing"]
        self.cells_xyz = List(state["cells_xyz"])
        self.cells_rgb = List(state["cells_rgb"])
        self.cells_classification = List(state["cells_classification"])
        self.cells_intensity = List(state["cells_intensity"])

    @property
    def max_key_value(self) -> int:
        return 1 << (
            2 * int(self.cell_count[0]).bit_length()
            + int(self.cell_count[2]).bit_length()
        )

    def insert(
        self,
        aabmin: npt.NDArray[np.float32],
        inv_aabb_size: npt.NDArray[np.float32],
        xyz: npt.NDArray[np.float32],
        rgb: npt.NDArray[np.uint8],
        classification: npt.NDArray[np.uint8],
        intensity: npt.NDArray[np.uint8],
        force: bool = False,
    ) -> tuple[
        npt.NDArray[np.float32],
        npt.NDArray[np.uint8],
        npt.NDArray[np.uint8],
        npt.NDArray[np.uint8],
        bool,
    ]:
        return _insert(  # type: ignore [no-any-return]
            self.cells_xyz,
            self.cells_rgb,
            self.cells_classification,
            self.cells_intensity,
            aabmin,
            inv_aabb_size,
            self.cell_count,
            xyz,
            rgb,
            classification,
            intensity,
            self.spacing,
            int(self.cell_count[0] - 1).bit_length(),
            force,
        )

    def needs_balance(self) -> bool:
        if self.cell_count[0] < 8:
            for cell in self.cells_xyz:
                if cell.shape[0] > 100000:
                    return True
        return False

    def balance(
        self,
        aabb_size: npt.NDArray[np.float32],
        aabmin: npt.NDArray[np.float32],
        inv_aabb_size: npt.NDArray[np.float32],
    ) -> None:
        t = aabb_size_to_subdivision_type(aabb_size)
        self.cell_count[0] += 1
        self.cell_count[1] += 1
        if t != SubdivisionType.QUADTREE:
            self.cell_count[2] += 1
        if self.cell_count[0] > 8:
            raise TilerException(
                f"The first value of the attribute cell count must be lower or equal to 8, "
                f"currently, it is {self.cell_count[0]}"
            )

        old_cells_xyz = self.cells_xyz
        old_cells_rgb = self.cells_rgb
        old_cells_classification = self.cells_classification
        old_cells_intensity = self.cells_intensity
        self.cells_xyz = List()
        self.cells_rgb = List()
        self.cells_classification = List()
        self.cells_intensity = List()
        for _ in range(self.max_key_value):
            self.cells_xyz.append(np.zeros((0, 3), dtype=np.float32))
            self.cells_rgb.append(np.zeros((0, 3), dtype=np.uint8))
            self.cells_classification.append(np.zeros((0, 1), dtype=np.uint8))
            self.cells_intensity.append(np.zeros((0, 1), dtype=np.uint8))

        for cellxyz, cellrgb, cellclassification, cellintensity in zip(
            old_cells_xyz, old_cells_rgb, old_cells_classification, old_cells_intensity
        ):
            self.insert(
                aabmin,
                inv_aabb_size,
                cellxyz,
                cellrgb,
                cellclassification,
                cellintensity,
                True,
            )

    def get_points(
        self, include_rgb: bool, include_classification: bool, include_intensity: bool
    ) -> npt.NDArray[np.uint8]:
        xyz = []
        rgb = []
        classification = []
        intensity = []
        for i in range(len(self.cells_xyz)):
            xyz.append(self.cells_xyz[i].view(np.uint8).ravel())
            if include_rgb:
                rgb.append(self.cells_rgb[i].ravel())
            if include_classification:
                classification.append(self.cells_classification[i].ravel())
            if include_intensity:
                intensity.append(self.cells_intensity[i].ravel())

        return np.concatenate((*xyz, *rgb, *classification, *intensity))

    def get_point_count(self) -> int:
        pt = 0
        for i in range(len(self.cells_xyz)):
            pt += self.cells_xyz[i].shape[0]
        return pt



import copy
from pathlib import Path

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from py3dtiles.exceptions import InvalidTilesetError, TilerException
from py3dtiles.tileset import BoundingVolumeBox, Tile
from py3dtiles.tileset.content import Pnts, PntsBody, PntsHeader


class TestTileContentManagement:
    def test_init_without_data(self, tmp_dir: Path) -> None:
        tile = Tile()

        assert tile.tile_content is None
        assert tile.content_uri is None
        with pytest.raises(
            RuntimeError, match="Cannot load a tile without a content_uri"
        ):
            tile.get_or_fetch_content(tmp_dir)

    def test_with_tile_content(self) -> None:
        tile = Tile()

        pnts = Pnts(PntsHeader(), PntsBody())
        tile.tile_content = pnts

        assert tile.get_or_fetch_content(None) == pnts
        assert tile.content_uri is None

    def test_with_path_content(self, tmp_dir_with_content: Path) -> None:
        tile_path = Path("r.pnts")
        tile = Tile(content_uri=tile_path)

        assert tile.tile_content is None
        assert isinstance(tile.get_or_fetch_content(tmp_dir_with_content), Pnts)
        assert isinstance(tile.tile_content, Pnts)

    def test_write(self, tmp_dir: Path) -> None:
        tile = Tile()
        pnts = Pnts(PntsHeader(), PntsBody())
        tile.tile_content = pnts

        with pytest.raises(
            TilerException, match="tile.content_uri is None, cannot write tile content"
        ):
            tile.write_content(tmp_dir)

        tile.content_uri = Path("rr.pnts")
        with pytest.raises(
            ValueError, match="No root_uri given and tile.content_uri is not absolute"
        ):
            tile.write_content(None)

        tile.write_content(tmp_dir)
        assert (tmp_dir / tile.content_uri).exists()

        tile.tile_content = None
        with pytest.raises(
            TilerException,
            match="The tile has no tile content. A tile content should be added in the tile.",
        ):
            tile.write_content(tmp_dir)


class TestTile:
    def test_constructor(self) -> None:
        tile = Tile()
        assert tile.bounding_volume is None
        assert tile.geometric_error == 500
        assert tile._refine == "ADD"
        assert tile.children == []
        assert_array_equal(tile.transform, np.identity(4))

        bounding_volume = BoundingVolumeBox()
        bounding_volume.set_from_list([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        tile = Tile(geometric_error=200, bounding_volume=bounding_volume)
        assert tile.bounding_volume == bounding_volume
        assert tile.geometric_error == 200
        assert tile._refine == "ADD"
        assert tile.children == []
        assert_array_equal(tile.transform, np.identity(4))

    def test_transform(self) -> None:
        tile = Tile()

        assert_array_equal(tile.transform, np.identity(4))

        # fmt: off
        tile.transform = np.array(
            [
                [1.0001, 0.0, 0.0, 0.0,],
                [0.0, 1.001, 0.0, 0.0,],
                [0.0, 0.0, 1.01, 0.0,],
                [0.0, 0.0, 0.0, 1.1,],
            ]
        )

        assert_array_equal(
            tile.transform,
            np.array(
                [
                    [1.0001, 0.0, 0.0, 0.0,],
                    [0.0, 1.001, 0.0, 0.0,],
                    [0.0, 0.0, 1.01, 0.0,],
                    [0.0, 0.0, 0.0, 1.1,],
                ]
            ),
        )
        # fmt: on

    def test_refine_mode(self) -> None:
        tile = Tile()
        assert tile.get_refine_mode() == "ADD"

        with pytest.raises(InvalidTilesetError):
            tile.set_refine_mode("replace")  # type: ignore [arg-type]

        tile.set_refine_mode("REPLACE")
        assert tile.get_refine_mode() == "REPLACE"

    def test_children(self) -> None:
        tile1 = Tile()

        assert tile1.children == []
        assert tile1.get_all_children() == []

        tile11 = Tile()
        tile1.add_child(tile11)
        assert tile1.children == [tile11]
        assert tile1.get_all_children() == [tile11]

        tile12 = Tile()
        tile1.add_child(tile12)
        assert tile1.children == [tile11, tile12]
        assert tile1.get_all_children() == [tile11, tile12]

        tile111 = Tile()
        tile11.add_child(tile111)

        assert tile1.children == [tile11, tile12]
        assert tile1.get_all_children() == [tile11, tile111, tile12]
        assert tile11.get_all_children() == [tile111]

    def test_to_dict(self) -> None:
        tile = Tile()

        with pytest.raises(InvalidTilesetError, match="Bounding volume is not set"):
            tile.to_dict()

        bounding_volume = BoundingVolumeBox()
        bounding_volume.set_from_list([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        tile.bounding_volume = bounding_volume

        # most simple case
        assert tile.to_dict() == {
            "boundingVolume": {
                "box": [
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    9.0,
                    10.0,
                    11.0,
                    12.0,
                ]
            },
            "geometricError": 500,
            "refine": "ADD",
        }

        tile.geometric_error = 3.14159
        tile.to_dict()  # just test if no error

        child_tile = Tile()
        child_tile.geometric_error = 21
        child_tile.bounding_volume = copy.deepcopy(bounding_volume)
        tile.add_child(child_tile)

        # cannot test now
        # child.set_content()

        assert tile.to_dict() == {
            "boundingVolume": {
                "box": [
                    1.0,
                    2.0,
                    3.0,
                    21.0,
                    0.0,
                    0.0,
                    0.0,
                    24.0,
                    0.0,
                    0.0,
                    0.0,
                    27.0,
                ]
            },
            "geometricError": 3.14159,
            "refine": "ADD",
            "children": [
                {
                    "boundingVolume": {
                        "box": [
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            8.0,
                            9.0,
                            10.0,
                            11.0,
                            12.0,
                        ]
                    },
                    "geometricError": 21,
                    "refine": "ADD",
                }
            ],
        }



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Generator, Generic, List, Optional, Tuple, TypeVar

from pyproj import CRS

from .shared_metadata import SharedMetadata
from .tiler_worker import TilerWorker

_SharedMetadataT = TypeVar("_SharedMetadataT", bound=SharedMetadata)
_TilerWorkerT = TypeVar("_TilerWorkerT", bound=TilerWorker[Any])


class Tiler(ABC, Generic[_SharedMetadataT, _TilerWorkerT]):
    """
    Tiler abstract class, this list of attributes and methods is used by convert.
    This class will organize the different tasks and their order of dispatch to the TilerWorker instances.

    You must set a name as class attribute and overwrite all abstract methods.

    Some methods are not required, overwrite them only if needed (like `validate_binary_data` or `memory_control`)
    """

    name = b""
    shared_metadata: _SharedMetadataT

    @abstractmethod
    def initialization(
        self,
        crs_out: Optional[CRS],
        working_dir: Path,
        number_of_jobs: int,
    ) -> None:
        """
        The __init__ method must only set attributes without any action.
        It is in this method that this work must be done (and the initialization of shared_metadata).

        The method will be called before all others.
        """

    @abstractmethod
    def get_worker(self) -> _TilerWorkerT:
        """
        Returns an instantiated tiler worker.
        """

    @abstractmethod
    def get_tasks(
        self, startup: float
    ) -> Generator[Tuple[bytes, List[bytes]], None, None]:
        """
        Yields tasks to be sent to workers.

        This methods will get called by the main convert function each time it wants new tasks to be fed to workers.
        Implementors should each time returns the task that has the biggest priority.


        """

    @abstractmethod
    def process_message(self, return_type: bytes, content: List[bytes]) -> bool:
        """
        Updates the state of the tiler in function of the return type and the returned data
        """

    @abstractmethod
    def write_tileset(self) -> None:
        """
        Writes the tileset file once the binary data written
        """

    def validate_binary_data(self) -> None:
        """
        Checks if the state of the tiler or the binary data written is correct.
        This method is called after the end of the conversion of this tiler (but before write_tileset)
        """

    def memory_control(self) -> None:
        """
        Method called at the end of each loop of the convert method.
        Checks if there is no too much memory used by the tiler and do actions in function
        """

    @abstractmethod
    def print_summary(self) -> None:
        """
        Prints the summary of the tiler before the start of the conversion.
        """

    def benchmark(self, benchmark_id: str, startup: float) -> None:
        """
        Prints benchmark info at the end of the conversion of this tiler and the writing of the tileset.
        """

    @abstractmethod
    def print_debug(
        self, now: float, number_of_jobs: int, number_of_idle_clients: int
    ) -> None:
        """
        Prints info about the progression of the conversion. Called everytime a tiler worker task is finished.
        """



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Literal

import numpy as np
import numpy.typing as npt

from .batch_table import BatchTable
from .feature_table import FeatureTable


class TileContent(ABC):
    header: TileContentHeader
    body: TileContentBody

    def to_array(self) -> npt.NDArray[np.uint8]:
        self.sync()
        header_arr = self.header.to_array()
        body_arr = self.body.to_array()
        return np.concatenate((header_arr, body_arr))

    def to_hex_str(self) -> str:
        arr = self.to_array()
        return " ".join(f"{x:02X}" for x in arr)

    def save_as(self, path: Path) -> None:
        tile_arr = self.to_array()
        with path.open("bw") as f:
            f.write(bytes(tile_arr))

    def __str__(self) -> str:
        return (
            "------ Tile header ------\n"
            + (str(self.header) if self.header is not None else "")
            + "\n"
            + "------ Tile body ------\n"
            + (str(self.body) if self.body is not None else "")
        )

    @abstractmethod
    def sync(self) -> None:
        """
        Allow to synchronize headers with contents.
        """

    @staticmethod
    @abstractmethod
    def from_array(array: npt.NDArray[np.uint8]) -> TileContent:
        ...


class TileContentHeader(ABC):
    magic_value: Literal[b"b3dm", b"pnts"]
    version: int

    def __init__(self) -> None:
        self.tile_byte_length = 0
        self.ft_json_byte_length = 0
        self.ft_bin_byte_length = 0
        self.bt_json_byte_length = 0
        self.bt_bin_byte_length = 0
        self.bt_length = 0  # number of models in the batch

    def __str__(self) -> str:
        infos = {
            "magic": self.magic_value,
            "version": self.version,
            "tile_byte_length": self.tile_byte_length,
            "json_feature_table_length": self.ft_json_byte_length,
            "bin_feature_table_length": self.ft_bin_byte_length,
            "json_batch_table_length": self.bt_json_byte_length,
            "bin_batch_table_length": self.bt_bin_byte_length,
        }
        return "\n".join(f"{key}: {value}" for key, value in infos.items())

    @staticmethod
    @abstractmethod
    def from_array(array: npt.NDArray[np.uint8]) -> TileContentHeader:
        ...

    @abstractmethod
    def to_array(self) -> npt.NDArray[np.uint8]:
        ...


class TileContentBody(ABC):
    batch_table: BatchTable
    feature_table: FeatureTable[Any, Any]

    @abstractmethod
    def __str__(self) -> str:
        ...

    @abstractmethod
    def to_array(self) -> npt.NDArray[np.uint8]:
        ...



from __future__ import annotations

from py3dtiles.typing import ExtensionDictType

registered_extensions: dict[str, type[BaseExtension]] = {}


class BaseExtension:
    """
    A base class to manage 3dtiles extension.

    If an extension is added somewhere in a tileset,
    the user must add the name of the extension in the attribute `extensions_used` of the class `TileSet`.
    Also, if the support of an extension is necessary to display the tileset,
    the name must be added in the attribute `extensions_required` of the class `TileSet`.
    """

    def __init__(self, name: str):
        self.name = name
        self.attributes: ExtensionDictType = {}

    @classmethod
    def from_dict(cls, extension_dict: ExtensionDictType) -> BaseExtension:
        """
        Creates an instance of the default extension class BaseExtension. This instance will store all attributes in a single field.

        :param extension_dict: a dict containing all attributes of the extension (keys and values).
        :return: a BaseExtension instance.
        """
        extension = cls("")
        extension.attributes = extension_dict
        return extension

    def to_dict(self) -> ExtensionDictType:
        """
        :return: a dict containing all attributes of the extension (keys and values).
        """
        return self.attributes


def create_extension(
    extension_key: str, extension_dict: ExtensionDictType
) -> BaseExtension:
    """
    Creates an instance of the extension if it is registered.
    Else, creates an instance of BaseExtension storing all attributes in a single field.

    :param extension_key: the name of the extension to create
    :param extension_dict: a dict containing all attributes of the extension (keys and values).
    :return: an extension.
    """
    extension_class = registered_extensions.get(extension_key, BaseExtension)
    return extension_class.from_dict(extension_dict)


def register_extension(
    extension_key: str, extension_class: type[BaseExtension]
) -> None:
    """
    Registers an extension by mapping the name of the extension with the corresponding class.

    :param extension_key: the name of the extension as a string.
    :param extension_class: the class of the extension.
    """
    registered_extensions[extension_key] = extension_class


def is_extension_registered(extension_key: str) -> bool:
    """
    Checks if the extension is registered.

    :param extension_key: the name of the extension as a string.
    :return: a boolean.
    """
    return extension_key in registered_extensions



import numpy as np
import pygltflib
from pytest import raises

from py3dtiles.tileset.content.gltf_utils import get_component_type_from_dtype


def test_component_type_from_dtype() -> None:
    assert get_component_type_from_dtype(np.dtype(np.int8)) == pygltflib.BYTE
    assert get_component_type_from_dtype(np.dtype(np.uint8)) == pygltflib.UNSIGNED_BYTE
    assert get_component_type_from_dtype(np.dtype(np.int16)) == pygltflib.SHORT
    assert (
        get_component_type_from_dtype(np.dtype(np.uint16)) == pygltflib.UNSIGNED_SHORT
    )
    assert get_component_type_from_dtype(np.dtype(np.uint32)) == pygltflib.UNSIGNED_INT
    assert get_component_type_from_dtype(np.dtype(np.float32)) == pygltflib.FLOAT
    with raises(
        ValueError,
        match="Cannot find a component type suitable for float64",
    ):
        get_component_type_from_dtype(np.dtype(np.float64))



import json
import unittest
from typing import List

import numpy as np
import numpy.typing as npt

from py3dtiles.exceptions import Invalid3dtilesError
from py3dtiles.tileset.content import PntsHeader
from py3dtiles.tileset.content.batch_table import (
    BatchTable,
    BatchTableBody,
    BatchTableHeader,
    ComponentNumpyType,
)
from py3dtiles.typing import BatchTableHeaderDataType


class TestBatchTableHeader(unittest.TestCase):
    def test_header_as_array_of_values(self) -> None:
        input_data: BatchTableHeaderDataType = {
            "id": ["unique id", "another unique id"],
            "displayName": ["Building name", "Another building name"],
            "yearBuilt": [1999, 2015],
            "address": [
                {"street": "Main Street", "houseNumber": "1"},
                {"street": "Main Street", "houseNumber": "2"},
            ],
        }
        bth = BatchTableHeader(input_data)

        if (
            json.loads(bth.to_array().tobytes().decode("utf-8")).items()
            != input_data.items()
        ):
            self.fail()


class TestBatchTableBody(unittest.TestCase):
    def test_empty_body(self) -> None:
        btb = BatchTableBody()
        np.testing.assert_equal(btb.to_array(), np.array([], np.uint8))

    def test_non_empty_body(self) -> None:
        btb = BatchTableBody(
            [
                np.array([1, 2, 3], np.uint8),
                np.array([4, 5, 6], np.uint8),
                np.array([7, 8, 9], np.uint8),
            ]
        )
        btb_array = btb.to_array()
        self.assertEqual(btb_array.nbytes, 16)
        np.testing.assert_equal(
            btb.to_array(),
            np.concatenate(
                (np.arange(1, 10, dtype=np.uint8), np.array([32] * 7, dtype=np.uint8))
            ),
        )


class TestBatchTable(unittest.TestCase):
    @staticmethod
    def get_batch_table_header_dict_only_binary() -> BatchTableHeaderDataType:
        return {
            "binProperty1": {
                "byteOffset": 0,
                "componentType": "UNSIGNED_BYTE",
                "type": "SCALAR",
            },
            "binProperty2": {
                "byteOffset": 4 * 1 * 1,
                "componentType": "INT",
                "type": "VEC2",
            },
        }

    @staticmethod
    def get_batch_table_header_dict_only_json() -> BatchTableHeaderDataType:
        return {
            "property1": [1, 2, 3, 4],
            "property2": [["a", "b"], ["a", "b"], ["a", "b"], ["a", "b"]],
        }

    def test_add_property_as_json(self) -> None:
        bt = BatchTable()
        bt.add_property_as_json("property_1", [1, 2, 3])
        self.assertTrue(set(bt.header.data.keys()) == {"property_1"})
        self.assertEqual(bt.header.data["property_1"], [1, 2, 3])
        self.assertEqual(bt.body.data, [])

    def test_add_property_as_binary(self) -> None:
        bt = BatchTable()

        bt.add_property_as_binary(
            "property_1", np.array([1, 2, 3], np.uint8), "UNSIGNED_BYTE", "SCALAR"
        )
        self.assertTrue(set(bt.header.data.keys()) == {"property_1"})
        self.assertEqual(
            bt.header.data["property_1"],
            {"byteOffset": 0, "componentType": "UNSIGNED_BYTE", "type": "SCALAR"},
        )
        self.assertEqual(len(bt.body.data), 1)
        np.testing.assert_equal(bt.body.data[0], np.array([1, 2, 3], np.uint8))

        bt.add_property_as_binary(
            "property_2", np.array([4, 5, 6], np.float32), "FLOAT", "SCALAR"
        )
        self.assertTrue(set(bt.header.data.keys()) == {"property_1", "property_2"})
        self.assertEqual(
            bt.header.data["property_2"],
            {"byteOffset": 3, "componentType": "FLOAT", "type": "SCALAR"},
        )
        self.assertEqual(len(bt.body.data), 2)
        np.testing.assert_equal(bt.body.data[1], np.array([4, 5, 6], np.float32))

    def test_to_array_with_empty_body(self) -> None:
        bt = BatchTable()
        bt.add_property_as_json("property_1", [1, 2, 3])
        self.assertTrue(np.array_equal(bt.to_array(), bt.header.to_array()))

    def test_to_array_with_non_empty_body(self) -> None:
        bt = BatchTable()
        bt.add_property_as_binary(
            "property_1", np.array([1, 2, 3], np.uint8), "UNSIGNED_BYTE", "SCALAR"
        )
        bt.add_property_as_binary(
            "property_2", np.array([4, 5, 6], np.uint8), "UNSIGNED_BYTE", "SCALAR"
        )
        self.assertTrue(
            np.array_equal(
                bt.to_array(),
                np.concatenate((bt.header.to_array(), bt.body.to_array())),
            )
        )

    def test_from_array_only_json(self) -> None:
        # build a minimal batch table
        batch_table_dict = self.get_batch_table_header_dict_only_json()
        batch_table_header = BatchTableHeader(batch_table_dict)
        batch_table_json = batch_table_header.to_array()

        # and tile content
        pnts_header = PntsHeader()
        pnts_header.bt_json_byte_length = len(batch_table_json)

        # import the array
        batch_table = BatchTable.from_array(pnts_header, batch_table_json)

        self.assertListEqual(batch_table.body.data, [])
        self.assertDictEqual(batch_table.header.data, batch_table_dict)

    def test_from_array_only_binary(self) -> None:
        # build a minimal batch table
        # json part
        batch_table_header_dict = self.get_batch_table_header_dict_only_binary()
        batch_table_header = BatchTableHeader(batch_table_header_dict)
        batch_table_json = batch_table_header.to_array()

        # binary part
        batch_table_body_binary: List[npt.NDArray[ComponentNumpyType]] = [
            np.array([1, 2, 3, 4], dtype=np.uint8),
            np.array([1, -1, 2, -2, 3, -3, 4, -4], dtype=np.int32),
        ]
        batch_table_body = BatchTableBody(batch_table_body_binary)

        # merge the 2 parts
        batch_table_array = np.concatenate(
            (batch_table_json, batch_table_body.to_array())
        )

        # and tile content
        pnts_header = PntsHeader()
        pnts_header.bt_json_byte_length = len(batch_table_json)

        batch_table = BatchTable.from_array(pnts_header, batch_table_array, 4)

        # import the array
        self.assertDictEqual(batch_table.header.data, batch_table_header_dict)

        self.assertEqual(len(batch_table.body.data), 2)
        np.testing.assert_equal(batch_table_body_binary[0], batch_table.body.data[0])
        np.testing.assert_equal(batch_table_body_binary[1], batch_table.body.data[1])

    def test_from_array_mix_of_json_and_binary(self) -> None:
        # build a minimal batch table
        # json part
        batch_table_header_dict = {
            **self.get_batch_table_header_dict_only_json(),
            **self.get_batch_table_header_dict_only_binary(),
        }
        batch_table_header = BatchTableHeader(batch_table_header_dict)

        # binary part
        batch_table_body_binary: List[npt.NDArray[ComponentNumpyType]] = [
            np.array([1, 2, 3, 4], dtype=np.uint8),
            np.array([1, -1, 2, -2, 3, -3, 4, -4], dtype=np.int32),
        ]
        batch_table_body = BatchTableBody(batch_table_body_binary)
        batch_table_json = batch_table_header.to_array()

        batch_table_array = np.concatenate(
            (batch_table_json, batch_table_body.to_array())
        )

        # and tile content
        pnts_header = PntsHeader()
        pnts_header.bt_json_byte_length = len(batch_table_json)

        # import the array
        batch_table = BatchTable.from_array(pnts_header, batch_table_array, 4)

        self.assertDictEqual(batch_table.header.data, batch_table_header_dict)

        self.assertEqual(len(batch_table.body.data), 2)
        np.testing.assert_equal(batch_table_body_binary[0], batch_table.body.data[0])
        np.testing.assert_equal(batch_table_body_binary[1], batch_table.body.data[1])

    def test_from_array_with_wrong_offset(self) -> None:
        # build a minimal batch table
        # json part
        batch_table_header_dict = self.get_batch_table_header_dict_only_binary()
        batch_table_header = BatchTableHeader(batch_table_header_dict)
        batch_table_json = batch_table_header.to_array()

        # binary part
        batch_table_body_binary: List[npt.NDArray[ComponentNumpyType]] = [
            np.array([1, 2, 3, 4], dtype=np.uint8),
            np.array([1, -1, 2, -2, 3, -3, 4, -4], dtype=np.int32),
        ]
        batch_table_body = BatchTableBody(batch_table_body_binary)

        # merge the 2 parts
        batch_table_array = np.concatenate(
            (batch_table_json, batch_table_body.to_array())
        )

        # and tile content
        pnts_header = PntsHeader()
        pnts_header.bt_json_byte_length = len(batch_table_json)

        # import the array (wrong batch_len)
        with self.assertRaisesRegex(
            Invalid3dtilesError,
            "The byte offset is 4 but the byte offset computed is 3",
        ):
            BatchTable.from_array(pnts_header, batch_table_array, 3)

        # import the array (wrong byteOffset)
        batch_table_header.data["binProperty2"]["byteOffset"] = 6  # type: ignore [call-overload]
        batch_table_json = batch_table_header.to_array()
        batch_table_array = np.concatenate(
            (batch_table_json, batch_table_body.to_array())
        )

        with self.assertRaisesRegex(
            Invalid3dtilesError,
            "The byte offset is 6 but the byte offset computed is 4",
        ):
            BatchTable.from_array(pnts_header, batch_table_array, 4)



from __future__ import annotations

from py3dtiles.tileset.extension import BaseExtension
from py3dtiles.typing import ExtensionDictType


class MockExtension(BaseExtension):
    @classmethod
    def from_dict(cls, extension_dict: ExtensionDictType) -> MockExtension:
        return cls("MockExtension")

    def to_dict(self) -> ExtensionDictType:
        return {}



import struct
import unittest
from filecmp import cmp
from pathlib import Path

import numpy as np
import pygltflib

from py3dtiles.exceptions import InvalidB3dmError
from py3dtiles.tilers.b3dm.wkb_utils import TriangleSoup
from py3dtiles.tileset.content import (
    B3dm,
    B3dmHeader,
    GltfAttribute,
    GltfPrimitive,
    read_binary_tile_content,
)
from py3dtiles.tileset.content.b3dm_feature_table import B3dmFeatureTable


class TestTileContentReader(unittest.TestCase):
    def test_read_and_write(self) -> None:
        tile_content = read_binary_tile_content(Path("tests/fixtures/buildings.b3dm"))
        if not isinstance(tile_content, B3dm):
            raise ValueError(
                f"The file 'tests/fixtures/buildings.b3dm' is a b3dm, not a {type(tile_content)}"
            )

        expected_tile_header_len = 28
        expected_feature_table_header_len = 20
        expected_feature_table_body_len = 0
        expected_batch_table_header_len = 64
        expected_batch_table_body_len = 0
        expected_gltf_header_len = 12  # magic + version + length
        expected_gltf_chunk_len = 8  # chunk length + chunk magic (JSON/BIN)
        expected_gltf_json_chunk_len = 1220
        expected_gltf_bin_chunk_len = 4968
        # Test feature table length
        self.assertEqual(
            tile_content.header.ft_json_byte_length, expected_feature_table_header_len
        )
        self.assertEqual(
            tile_content.header.ft_bin_byte_length, expected_feature_table_body_len
        )
        # Test batch table length
        self.assertEqual(
            tile_content.header.bt_json_byte_length, expected_batch_table_header_len
        )
        self.assertEqual(
            tile_content.header.bt_bin_byte_length, expected_batch_table_body_len
        )
        # Test gltf length
        (
            _,
            _,
            length,
            json_blob_length,
            _,
            json_blob,
            bin_blob_length,
            _,
            bin_blob,
        ) = tile_content.body.gltf.save_to_bytes()
        self.assertEqual(
            struct.unpack("<I", json_blob_length)[0], expected_gltf_json_chunk_len
        )
        self.assertEqual(len(json_blob), expected_gltf_json_chunk_len)
        self.assertEqual(
            struct.unpack("<I", bin_blob_length)[0], expected_gltf_bin_chunk_len
        )
        self.assertEqual(len(bin_blob), expected_gltf_bin_chunk_len)
        # Test total length
        self.assertEqual(
            tile_content.header.tile_byte_length,
            expected_tile_header_len
            + expected_feature_table_header_len
            + expected_feature_table_body_len
            + expected_batch_table_header_len
            + expected_batch_table_body_len
            + expected_gltf_header_len
            + expected_gltf_chunk_len
            + expected_gltf_json_chunk_len
            + expected_gltf_chunk_len
            + expected_gltf_bin_chunk_len,
        )
        self.assertDictEqual(
            tile_content.body.batch_table.header.data,
            {"id": ["BATIMENT0000000240853073", "BATIMENT0000000240853157"]},
        )
        self.assertEqual(tile_content.body.feature_table.get_batch_length(), 2)
        self.assertEqual(tile_content.body.gltf.asset.version, "2.0")
        self.assertEqual(len(b"".join(tile_content.body.gltf.save_to_bytes())), 6216)

        path_name = Path("tests/output_tests/buildings.b3dm")
        path_name.parent.mkdir(parents=True, exist_ok=True)
        tile_content.save_as(path_name)
        self.assertTrue(cmp("tests/fixtures/buildings.b3dm", path_name))

    def test_str(self) -> None:
        tile_content = read_binary_tile_content(Path("tests/fixtures/buildings.b3dm"))
        expected_string_components = [
            "------ Tile header ------",
            "magic: b'b3dm'",
            "version: 1",
            "tile_byte_length: 6328",
            "json_feature_table_length: 20",
            "bin_feature_table_length: 0",
            "json_batch_table_length: 64",
            "bin_batch_table_length: 0",
            "------ Tile body ------",
            "feature_table_batch_length: 2",
            "gltf_magic: b'glTF'",
            "gltf_version: 2.0",
            "gltf_length: 6216",
            "gltf_json_chunk_length: 1220",
            "gltf_bin_chunk_length: 4968",
        ]
        string_components = str(tile_content).split("\n")
        for expected_line, line in zip(expected_string_components, string_components):
            assert expected_line == line


class TestTileContentBuilder(unittest.TestCase):
    def test_build(self) -> None:
        with open("tests/fixtures/building/building.wkb", "rb") as f:
            wkb = f.read()
        ts = TriangleSoup.from_wkb_multipolygon(wkb)

        # translation : 1842015.125, 5177109.25, 247.87364196777344
        transform = np.array(
            [
                [1, 0, 0, 1842015.125],
                [0, 1, 0, 5177109.25],
                [0, 0, 1, 247.87364196777344],
                [0, 0, 0, 1],
            ],
            dtype=float,
        )
        feature_table = B3dmFeatureTable()
        feature_table.set_batch_length(1)
        t = B3dm.from_numpy_arrays(
            ts.vertices,
            ts.triangle_indices,
            feature_table=feature_table,
            normal=ts.compute_normals(),
            transform=transform,
        )

        # get an array
        t.to_array()
        self.assertEqual(t.header.version, 1.0)

        # Test the tile byte length
        expected_tile_header_len = 28
        expected_feature_table_header_len = 20
        expected_feature_table_body_len = 0
        expected_batch_table_header_len = 0
        expected_batch_table_body_len = 0
        expected_gltf_header_len = 12  # magic + version + length
        expected_gltf_chunk_len = 8  # chunk length + chunk magic (JSON/BIN)
        expected_gltf_json_chunk_len = 1204
        expected_gltf_bin_chunk_len = 1104
        # Test feature table length
        self.assertEqual(
            t.header.ft_json_byte_length, expected_feature_table_header_len
        )
        self.assertEqual(t.header.ft_bin_byte_length, expected_feature_table_body_len)
        # Test batch table length
        self.assertEqual(t.header.bt_json_byte_length, expected_batch_table_header_len)
        self.assertEqual(t.header.bt_bin_byte_length, expected_batch_table_body_len)
        # Test gltf length
        (
            _,
            _,
            length,
            json_blob_length,
            _,
            json_blob,
            bin_blob_length,
            _,
            bin_blob,
        ) = t.body.gltf.save_to_bytes()
        self.assertEqual(
            struct.unpack("<I", json_blob_length)[0], expected_gltf_json_chunk_len
        )
        self.assertEqual(len(json_blob), expected_gltf_json_chunk_len)
        self.assertEqual(
            struct.unpack("<I", bin_blob_length)[0], expected_gltf_bin_chunk_len
        )
        self.assertEqual(len(bin_blob), expected_gltf_bin_chunk_len)
        # Test total length
        self.assertEqual(
            t.header.tile_byte_length,
            expected_tile_header_len
            + expected_feature_table_header_len
            + expected_feature_table_body_len
            + expected_batch_table_header_len
            + expected_batch_table_body_len
            + expected_gltf_header_len
            + expected_gltf_chunk_len
            + expected_gltf_json_chunk_len
            + expected_gltf_chunk_len
            + expected_gltf_bin_chunk_len,
        )
        self.assertEqual(
            t.header.tile_byte_length % 8, 0
        )  # tile bytes must be 8-byte aligned

        # Test the feature table byte lengths
        json_feature_table_end = B3dmHeader.BYTE_LENGTH + t.header.ft_json_byte_length
        self.assertEqual(json_feature_table_end % 8, 0)

        # This 20 corresponds to the length of the JSON feature table ("{'BATCH_LENGTH':1}") + the padding to align on 8-bytes
        self.assertEqual(t.header.ft_json_byte_length, 20)

        bin_feature_table_end = json_feature_table_end + t.header.ft_bin_byte_length
        self.assertEqual(bin_feature_table_end % 8, 0)
        self.assertEqual(t.header.ft_bin_byte_length, 0)

        # Test the batch table byte lengths
        self.assertEqual(t.header.bt_json_byte_length, 0)
        self.assertEqual(t.header.bt_bin_byte_length, 0)

        # Test the gltf byte length
        gltf_start_bounding = (
            bin_feature_table_end
            + t.header.bt_json_byte_length
            + t.header.bt_bin_byte_length
        )
        self.assertEqual(
            gltf_start_bounding % 8, 0
        )  # the gltf part must be 8-byte aligned
        self.assertEqual(
            len(b"".join(t.body.gltf.save_to_bytes())) % 8, 0
        )  # gltf bytes must be 8-byte aligned

        # t.save_as("/tmp/py3dtiles_test_build_1.b3dm")


class TestTexturedTileBuilder(unittest.TestCase):
    def test_build(self) -> None:
        with open("tests/fixtures/square.wkb", "rb") as f:
            wkb = f.read()
        with open("tests/fixtures/squareUV.wkb", "rb") as f:
            wkbuv = f.read()
        ts = TriangleSoup.from_wkb_multipolygon(wkb, [wkbuv])

        t = B3dm.from_numpy_arrays(
            ts.vertices,
            ts.triangle_indices,
            normal=ts.compute_normals(),
            uvs=ts.get_data(0),
            texture_uri="tests/fixtures/squaretexture.jpg",
        )

        # get an array
        t.to_array()
        self.assertEqual(t.header.version, 1.0)
        self.assertEqual(t.header.tile_byte_length, 1692)
        self.assertEqual(t.header.ft_json_byte_length, 0)
        self.assertEqual(t.header.ft_bin_byte_length, 0)
        self.assertEqual(t.header.bt_json_byte_length, 0)
        self.assertEqual(t.header.bt_bin_byte_length, 0)
        accessors = t.body.gltf.accessors
        self.assertEqual(accessors[1].min, [0, 0, 0])
        self.assertEqual(accessors[1].max, [10, 10, 0])

        t_without_normals = B3dm.from_numpy_arrays(
            ts.vertices,
            ts.triangle_indices,
            uvs=ts.get_data(0),
            texture_uri="tests/fixtures/squaretexture.jpg",
        )

        # get an array
        t_without_normals.to_array()
        self.assertEqual(t_without_normals.header.tile_byte_length, 1452)

        # t.save_as("/tmp/py3dtiles_test_build_1.b3dm")

    def test_build_batchids(self) -> None:
        with open("tests/fixtures/square.wkb", "rb") as f:
            wkb = f.read()
        ts = TriangleSoup.from_wkb_multipolygon(wkb)
        ft = B3dmFeatureTable()
        ft.set_batch_length(2)

        t = B3dm.from_numpy_arrays(
            ts.vertices,
            ts.triangle_indices,
            feature_table=ft,
            normal=np.array(
                [
                    [0.0, 0.0, -1.0],
                    [0.0, 0.0, -1.0],
                    [0.0, 0.0, -1.0],
                    [-0.0, -0.0, -1.0],
                    [-0.0, -0.0, -1.0],
                    [-0.0, -0.0, -1.0],
                ],
                dtype=np.float32,
            ),
            batchids=np.array([0, 0, 0, 1, 1, 1], dtype=np.uint32),
        )

        # get an array
        t.to_array()
        self.assertEqual(t.header.version, 1.0)
        self.assertEqual(t.header.tile_byte_length, 1480)
        self.assertEqual(t.header.ft_json_byte_length, 20)
        self.assertEqual(t.header.ft_bin_byte_length, 0)
        self.assertEqual(t.header.bt_json_byte_length, 0)
        self.assertEqual(t.header.bt_bin_byte_length, 0)

    def test_build_multiple_parts(self) -> None:
        with open("tests/fixtures/square.wkb", "rb") as f:
            wkb = f.read()
        with open("tests/fixtures/squareUV.wkb", "rb") as f:
            wkbuv = f.read()
        ts_0 = TriangleSoup.from_wkb_multipolygon(wkb)
        ts_1 = TriangleSoup.from_wkb_multipolygon(wkb, [wkbuv])
        for triangle in ts_1.triangles[0]:
            for point in triangle:
                point[1] += 30
        ts_2 = TriangleSoup.from_wkb_multipolygon(wkb)
        for triangle in ts_2.triangles[0]:
            for point in triangle:
                point[1] -= 30
        ts_3 = TriangleSoup.from_wkb_multipolygon(wkb)
        for triangle in ts_3.triangles[0]:
            for point in triangle:
                point[1] += 60
        ts_4 = TriangleSoup.from_wkb_multipolygon(wkb, [wkbuv])
        for triangle in ts_4.triangles[0]:
            for point in triangle:
                point[1] += 90
        vertex_color = GltfAttribute(
            "COLOR_0",
            pygltflib.VEC3,
            pygltflib.FLOAT,
            np.array(
                [
                    [0, 0, 1],
                    [1, 0, 0],
                    [0, 1, 0],
                    [1, 0, 0],
                    [0, 1, 0],
                    [0, 0, 1],
                ],
                dtype=np.float32,
            ),
        )
        material = pygltflib.Material(
            pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                baseColorFactor=[1.0, 0.0, 0.0, 0.75],
                metallicFactor=0.5,
                roughnessFactor=0.5,
            )
        )
        ft = B3dmFeatureTable()
        ft.set_batch_length(0)
        primitives = [
            GltfPrimitive(ts_0.vertices, ts_0.triangle_indices, ts_0.compute_normals()),
            GltfPrimitive(
                ts_1.vertices,
                ts_1.triangle_indices,
                ts_1.compute_normals(),
                ts_1.get_data(0),
                texture_uri="squaretexture.jpg",
            ),
            GltfPrimitive(
                ts_2.vertices,
                ts_2.triangle_indices,
                ts_2.compute_normals(),
                additional_attributes=[vertex_color],
            ),
            GltfPrimitive(
                ts_3.vertices,
                ts_3.triangle_indices,
                ts_3.compute_normals(),
                material=material,
            ),
            GltfPrimitive(
                np.array(ts_4.triangles[0]).flatten().reshape((-1, 3)),
                normals=ts_4.compute_normals(),
                uvs=ts_4.get_data(0),
                texture_uri="squaretexture.jpg",
                material=material,
            ),
        ]

        t = B3dm.from_primitives(
            primitives,
            feature_table=ft,
        )

        # get an array
        t.to_array()
        self.assertEqual(t.header.version, 1.0)
        self.assertEqual(t.header.tile_byte_length, 5776)
        self.assertEqual(t.header.ft_json_byte_length, 20)
        self.assertEqual(t.header.ft_bin_byte_length, 0)
        self.assertEqual(t.header.bt_json_byte_length, 0)
        self.assertEqual(t.header.bt_bin_byte_length, 0)
        accessors = t.body.gltf.accessors
        gltf_primitives = t.body.gltf.meshes[0].primitives
        self.assertEqual(accessors[1].min, [0, 0, 0])
        self.assertEqual(accessors[1].max, [10, 10, 0])
        self.assertEqual(len(accessors), 17)
        self.assertEqual(len(gltf_primitives), 5)
        self.assertEqual(gltf_primitives[2].attributes.COLOR_0, 10)
        self.assertEqual(
            t.body.gltf.materials[3].pbrMetallicRoughness.baseColorFactor,
            [1.0, 0.0, 0.0, 0.75],
        )
        with self.assertRaises(
            InvalidB3dmError
        ):  # Assert an error is raised if a GltfPrimitive as UVs but no texture URI
            _ = B3dm.from_primitives(
                [
                    GltfPrimitive(
                        ts_1.vertices,
                        ts_1.triangle_indices,
                        uvs=ts_1.get_data(0),
                        material=material,
                    ),
                ]
            )

    def test_build_and_read_gltf_content(self) -> None:
        """See "Create a mesh, convert to bytes, convert back to mesh" section from
        https://gitlab.com/dodgyville/pygltflib/. This test is an adaptation of the pygltflib
        documentation.

        """
        with open("tests/fixtures/building/building.wkb", "rb") as f:
            wkb = f.read()
        ts = TriangleSoup.from_wkb_multipolygon(wkb)
        t = B3dm.from_numpy_arrays(
            ts.vertices,
            ts.triangle_indices,
            normal=ts.compute_normals(),
        )

        gltf = t.body.gltf
        gltf_binary_blob = gltf.binary_blob()

        triangles_accessor = gltf.accessors[gltf.meshes[0].primitives[0].indices]
        triangles_buffer_view = gltf.bufferViews[triangles_accessor.bufferView]
        triangles = np.frombuffer(
            gltf_binary_blob[
                triangles_buffer_view.byteOffset
                + triangles_accessor.byteOffset : triangles_buffer_view.byteOffset
                + triangles_buffer_view.byteLength
            ],
            dtype="uint8",
            count=triangles_accessor.count,
        ).reshape(int(triangles_accessor.count / 3), 3)
        np.testing.assert_array_equal(triangles, ts.triangle_indices)

        points_accessor = gltf.accessors[
            gltf.meshes[0].primitives[0].attributes.POSITION
        ]
        points_buffer_view = gltf.bufferViews[points_accessor.bufferView]
        points = np.frombuffer(
            gltf_binary_blob[
                points_buffer_view.byteOffset
                + points_accessor.byteOffset : points_buffer_view.byteOffset
                + points_buffer_view.byteLength
            ],
            dtype="float32",
            count=points_accessor.count * 3,
        ).reshape(points_accessor.count, 3)
        np.testing.assert_array_equal(points, ts.vertices)

        normals_accessor = gltf.accessors[
            gltf.meshes[0].primitives[0].attributes.NORMAL
        ]
        normals_buffer_view = gltf.bufferViews[normals_accessor.bufferView]
        normals = np.frombuffer(
            gltf_binary_blob[
                normals_buffer_view.byteOffset
                + normals_accessor.byteOffset : normals_buffer_view.byteOffset
                + normals_buffer_view.byteLength
            ],
            dtype="float32",
            count=normals_accessor.count * 3,
        ).reshape(normals_accessor.count, 3)
        np.testing.assert_array_equal(normals, ts.compute_normals())



import json
from pathlib import Path

from py3dtiles.export import from_directory

DATA_DIRECTORY = Path(__file__).parent / "fixtures"


def test_export(tmp_dir: Path) -> None:
    from_directory(DATA_DIRECTORY / "building", offset=None, output_dir=tmp_dir)

    # basic asserts
    tileset_path = tmp_dir / "tileset.json"
    with tileset_path.open() as f:
        tileset = json.load(f)

    expecting_box = [0.028, 2.054, -0.028, 8.776, 0, 0, 0, 0.0, 0, 0, 0, 7.327]
    box = [round(value, 4) for value in tileset["root"]["boundingVolume"]["box"]]
    assert box == expecting_box

    assert Path(tmp_dir, "tiles", "1.b3dm").exists()



import os
import pickle
import struct
import time
from pathlib import PurePath
from typing import List

import zmq

from py3dtiles.tilers.base_tiler import TilerWorker
from py3dtiles.utils import READER_MAP

from .node import NodeCatalog, NodeProcess
from .pnts import pnts_writer
from .point_message_type import PointManagerMessage, PointWorkerMessageType
from .point_shared_metadata import PointSharedMetadata


class PointTilerWorker(TilerWorker[PointSharedMetadata]):
    def execute(
        self, skt: zmq.Socket[bytes], command: bytes, content: List[bytes]
    ) -> None:
        if command == PointManagerMessage.READ_FILE.value:
            self.execute_read_file(skt, content)
        elif command == PointManagerMessage.PROCESS_JOBS.value:
            self.execute_process_jobs(skt, content)
        elif command == PointManagerMessage.WRITE_PNTS.value:
            self.execute_write_pnts(skt, content[1], content[0])
        else:
            raise NotImplementedError(f"Unknown command {command!r}")

    def execute_read_file(self, skt: zmq.Socket[bytes], content: List[bytes]) -> None:
        parameters = pickle.loads(content[0])

        extension = PurePath(parameters["filename"]).suffix
        if extension in READER_MAP:
            reader = READER_MAP[extension]
        else:
            raise ValueError(
                f"The file with {extension} extension can't be read, "
                f"the available extensions are: {READER_MAP.keys()}"
            )

        reader_gen = reader.run(
            parameters["filename"],
            parameters["offset_scale"],
            parameters["portion"],
            self.shared_metadata.transformer,
            self.shared_metadata.color_scale,
            self.shared_metadata.write_intensity,
        )
        for coords, colors, classification, intensity in reader_gen:
            skt.send_multipart(
                [
                    PointWorkerMessageType.NEW_TASK.value,
                    b"",
                    pickle.dumps(
                        {
                            "xyz": coords,
                            "rgb": colors,
                            "classification": classification,
                            "intensity": intensity,
                        }
                    ),
                    struct.pack(">I", len(coords)),
                ],
                copy=False,
            )

        skt.send_multipart([PointWorkerMessageType.READ.value])

    def execute_write_pnts(
        self, skt: zmq.Socket[bytes], content: bytes, node_name: bytes
    ) -> None:
        pnts_writer_gen = pnts_writer.run(
            content,
            self.shared_metadata.out_folder,
            self.shared_metadata.write_rgb,
            self.shared_metadata.write_classification,
            self.shared_metadata.write_intensity,
        )
        for total in pnts_writer_gen:
            skt.send_multipart(
                [
                    PointWorkerMessageType.PNTS_WRITTEN.value,
                    struct.pack(">I", total),
                    node_name,
                ]
            )

    def execute_process_jobs(
        self, skt: zmq.Socket[bytes], content: List[bytes]
    ) -> None:
        begin = time.time()
        log_enabled = self.shared_metadata.verbosity >= 2
        if log_enabled:
            log_filename = f"py3dtiles-{os.getpid()}.log"
            log_file = open(log_filename, "a")
        else:
            log_file = None

        i = 0
        while i < len(content):
            name = content[i]
            node = content[i + 1]
            count = struct.unpack(">I", content[i + 2])[0]
            tasks = content[i + 3 : i + 3 + count]
            i += 3 + count

            node_catalog = NodeCatalog(
                node,
                name,
                self.shared_metadata.root_aabb,
                self.shared_metadata.root_spacing,
            )

            node_process = NodeProcess(
                node_catalog,
                self.shared_metadata.scale[0],
                name,
                tasks,
                begin,
                log_file,
            )
            for proc_name, proc_data, proc_point_count in node_process.run():
                skt.send_multipart(
                    [
                        PointWorkerMessageType.NEW_TASK.value,
                        proc_name,
                        proc_data,
                        struct.pack(">I", proc_point_count),
                    ],
                    copy=False,
                    block=False,
                )

            if log_enabled:
                print(f"save on disk {name!r} [{time.time() - begin}]", file=log_file)

            # save node state on disk
            if len(name) > 0:
                data = node_catalog.dump(name, node_process.infer_depth_from_name() - 1)
            else:
                data = b""

            if log_enabled:
                print(f"saved on disk [{time.time() - begin}]", file=log_file)

            skt.send_multipart(
                [
                    PointWorkerMessageType.PROCESSED.value,
                    pickle.dumps(
                        {
                            "name": name,
                            "total": node_process.total_point_count,
                            "data": data,
                        }
                    ),
                ],
                copy=False,
            )

        if log_enabled:
            print(
                "[<] return result [{} sec] [{}]".format(
                    round(time.time() - begin, 2), time.time() - begin
                ),
                file=log_file,
                flush=True,
            )
            if log_file is not None:
                log_file.close()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from py3dtiles.tileset.extension import create_extension
from py3dtiles.typing import ExtraDictType, RootPropertyDictType

if TYPE_CHECKING:
    from typing_extensions import Self

    from py3dtiles.tileset.extension import BaseExtension

_JsonDictT = TypeVar("_JsonDictT", bound=RootPropertyDictType)


class RootProperty(ABC, Generic[_JsonDictT]):
    """
    One the 3DTiles notions defined as an abstract data model through
    a schema of the 3DTiles specifications (either core of extensions).
    """

    def __init__(self) -> None:
        self.extensions: dict[str, BaseExtension] = {}
        self.extras: ExtraDictType = {}

    @classmethod
    @abstractmethod
    def from_dict(cls, data_dict: _JsonDictT) -> Self:
        ...

    @abstractmethod
    def to_dict(self) -> _JsonDictT:
        ...

    def add_root_properties_to_dict(self, dict_data: _JsonDictT) -> _JsonDictT:
        # we cannot merge root_property_data without mypy issues
        if self.extensions:
            dict_data["extensions"] = {
                name: extension.to_dict() for name, extension in self.extensions.items()
            }

        if self.extras:
            dict_data["extras"] = self.extras

        return dict_data

    def set_properties_from_dict(
        self,
        dict_data: _JsonDictT,
    ) -> None:
        if "extensions" in dict_data:
            for key in dict_data["extensions"]:
                self.extensions[key] = create_extension(
                    key, dict_data["extensions"][key]
                )
                if not self.extensions[key].name:
                    self.extensions[key].name = key
        if "extras" in dict_data:
            self.extras = dict_data["extras"]



from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar

import zmq

from py3dtiles.tilers.base_tiler.shared_metadata import SharedMetadata

_SharedMetadataT = TypeVar("_SharedMetadataT", bound=SharedMetadata)


class TilerWorker(ABC, Generic[_SharedMetadataT]):
    def __init__(self, shared_metadata: _SharedMetadataT):
        # The attribute shared_metadata must not be modified by any tiler worker
        self.shared_metadata = shared_metadata

    @abstractmethod
    def execute(
        self, skt: zmq.Socket[bytes], command: bytes, content: List[bytes]
    ) -> None:
        """
        Executes a command sent by the tiler. The method returns directly the response with the skt variable.
        """



from pathlib import Path

import numpy as np
from numpy.testing import assert_array_equal
from pytest_benchmark.fixture import BenchmarkFixture

from py3dtiles.tilers.point.node import Grid, Node
from py3dtiles.tilers.point.node.distance import is_point_far_enough
from py3dtiles.utils import node_name_to_path

# test point
xyz = np.array([0.25, 0.25, 0.25], dtype=np.float32)
to_insert = np.array([[0.25, 0.25, 0.25]], dtype=np.float32)
xyz2 = np.array([0.6, 0.6, 0.6], dtype=np.float32)
rgb = np.zeros((1, 3), dtype=np.uint8)
classification = np.zeros((1, 1), dtype=np.uint8)
intensity = np.zeros((1, 1), dtype=np.uint8)
sample_points = np.array(
    [[x / 30, x / 30, x / 30] for x in range(30)], dtype=np.float32
)


def test_grid_insert(grid: Grid, node: Node) -> None:
    assert (
        grid.insert(
            node.aabb[0], node.inv_aabb_size, to_insert, rgb, classification, intensity
        )[0].shape[0]
        == 0
    )
    assert (
        grid.insert(
            node.aabb[0], node.inv_aabb_size, to_insert, rgb, classification, intensity
        )[0].shape[0]
        == 1
    )


def test_grid_insert_perf(grid: Grid, node: Node, benchmark: BenchmarkFixture) -> None:
    benchmark(
        grid.insert,
        node.aabb[0],
        node.inv_aabb_size,
        to_insert,
        rgb,
        classification,
        intensity,
    )


def test_grid_getpoints(grid: Grid, node: Node) -> None:
    grid.insert(
        node.aabb[0], node.inv_aabb_size, to_insert, rgb, classification, intensity
    )
    points = grid.get_points(True, True, True)
    ref = np.hstack([to_insert.view(np.uint8), rgb, classification, intensity])[0]
    assert_array_equal(points, ref)


def test_grid_getpoints_perf(
    grid: Grid, node: Node, benchmark: BenchmarkFixture
) -> None:
    assert (
        grid.insert(
            node.aabb[0], node.inv_aabb_size, to_insert, rgb, classification, intensity
        )[0].shape[0]
        == 0
    )
    benchmark(grid.get_points, True, True, True)


def test_grid_get_point_count(grid: Grid, node: Node) -> None:
    grid.insert(
        node.aabb[0], node.inv_aabb_size, to_insert, rgb, classification, intensity
    )
    assert len(grid.get_points(False, False, False)) == 1 * (3 * 4) * 1
    grid.insert(
        node.aabb[0], node.inv_aabb_size, to_insert, rgb, classification, intensity
    )
    assert len(grid.get_points(False, False, False)) == 1 * (3 * 4) * 1


def test_is_point_far_enough() -> None:
    points = np.array(
        [
            [1, 1, 1],
            [0.2, 0.2, 0.2],
            [0.4, 0.4, 0.4],
        ],
        dtype=np.float32,
    )
    assert not is_point_far_enough(points, xyz, 0.25**2)
    assert is_point_far_enough(points, xyz2, 0.25**2)


def test_is_point_far_enough_perf(benchmark: BenchmarkFixture) -> None:
    benchmark(is_point_far_enough, sample_points, xyz, 0.25**2)


def test_short_name_to_path() -> None:
    short_tile_name = b""
    path = node_name_to_path(Path("work"), short_tile_name)
    assert path == Path("work/r")


def test_long_name_to_path() -> None:
    long_tile_name = b"110542453782"
    path = node_name_to_path(Path("work"), long_tile_name)
    assert path == Path("work/11054245/r3782")


def test_long_name_to_path_with_extension() -> None:
    long_tile_name = b"110542453782"
    path = node_name_to_path(Path("work"), long_tile_name, suffix=".pnts")
    assert path == Path("work/11054245/r3782.pnts")


def test_long_name_to_path_with_short_split() -> None:
    long_tile_name = b"110542453782"
    path = node_name_to_path(Path("work"), long_tile_name, split_len=2)
    assert path == Path("work/11/05/42/45/37/r82")



import unittest
from pathlib import Path

import numpy as np
from numpy.testing import assert_array_equal

from py3dtiles.tileset.content import Pnts, PntsHeader, read_binary_tile_content
from py3dtiles.tileset.content.pnts_feature_table import (
    PntsFeatureTableHeader,
    SemanticPoint,
)


class TestTileContentReader(unittest.TestCase):
    def test_read(self) -> None:
        tile = Pnts.from_file(Path("tests/fixtures/pointCloudRGB.pnts"))

        self.assertEqual(tile.header.version, 1.0)
        self.assertEqual(tile.header.tile_byte_length, 15176)
        self.assertEqual(tile.header.ft_json_byte_length, 148)
        self.assertEqual(tile.header.ft_bin_byte_length, 15000)
        self.assertEqual(tile.header.bt_json_byte_length, 0)
        self.assertEqual(tile.header.bt_bin_byte_length, 0)

        feature_table = tile.body.feature_table
        pt_color = feature_table.get_feature_color_at(0)
        if pt_color is None:
            raise RuntimeError("pt_color should not be None")
        assert_array_equal(
            np.array([44, 243, 209], dtype=np.uint8),
            pt_color,
        )

    def test_str(self) -> None:
        tile_content = read_binary_tile_content(
            Path("tests/fixtures/pointCloudRGB.pnts")
        )
        expected_string_components = [
            "------ Tile header ------",
            "magic: b'pnts'",
            "version: 1",
            "tile_byte_length: 15176",
            "json_feature_table_length: 148",
            "bin_feature_table_length: 15000",
            "json_batch_table_length: 0",
            "bin_batch_table_length: 0",
            "------ Tile body ------",
            "feature_table_header: {'"
            "POINTS_LENGTH': 1000"
            ", 'RTC_CENTER': [1215012.8828876738, -4736313.051199594, 4081605.22126042]"
            ", 'POSITION': {'byteOffset': 0}, 'RGB': {'byteOffset': 12000}}",
            "points_length: 1000",
            "first_point_position: [ 2.19396     4.489685   -0.17107764]",
            "first_point_color: [ 44 243 209]",
            "first_point_normal: None",
        ]
        string_components = str(tile_content).split("\n")
        for expected_line, line in zip(expected_string_components, string_components):
            assert expected_line == line


class TestTileBuilder(unittest.TestCase):
    def test_build_without_colors(self) -> None:
        tread = Pnts.from_file(Path("tests/fixtures/pointCloudRGB.pnts"))
        feature_0_position = tread.body.feature_table.get_feature_position_at(0)

        # create features
        positions = []
        for i in range(tread.body.feature_table.header.points_length):
            feature_position = tread.body.feature_table.get_feature_position_at(i)
            positions.append(feature_position)
        position_array = np.array(positions).flatten()

        # create a tile
        feature_table_header = PntsFeatureTableHeader.from_semantic(
            SemanticPoint.POSITION,
            None,
            None,
            len(positions),
        )
        t = Pnts.from_features(feature_table_header, position_array)

        # configure the tile
        rtc = (1215012.8828876738, -4736313.051199594, 4081605.22126042)
        t.body.feature_table.header.rtc = rtc

        # get an array
        tile_arr = t.to_array()
        t2 = Pnts.from_array(tile_arr)
        self.assertEqual(t2.header.version, 1.0)

        # Test the tile byte length
        self.assertEqual(t2.header.tile_byte_length, 12152)
        self.assertEqual(
            t2.header.tile_byte_length % 8, 0
        )  # tile bytes must be 8-byte aligned

        # Test the feature table byte lengths
        json_feature_table_end = PntsHeader.BYTE_LENGTH + t2.header.ft_json_byte_length
        self.assertEqual(json_feature_table_end % 8, 0)
        self.assertEqual(t2.header.ft_json_byte_length, 124)
        bin_feature_table_end = json_feature_table_end + t2.header.ft_bin_byte_length
        self.assertEqual(bin_feature_table_end % 8, 0)
        self.assertEqual(t2.header.ft_bin_byte_length, 12000)

        self.assertEqual(t2.header.bt_json_byte_length, 0)
        self.assertEqual(t2.header.bt_bin_byte_length, 0)

        feature_table = t.body.feature_table

        assert_array_equal(feature_0_position, feature_table.get_feature_position_at(0))

    def test_build(self) -> None:
        tread = Pnts.from_file(Path("tests/fixtures/pointCloudRGB.pnts"))

        # create features
        positions = []
        colors = []
        for i in range(tread.body.feature_table.header.points_length):
            (
                feature_position,
                feature_color,
                _,
            ) = tread.body.feature_table.get_feature_at(i)
            positions.append(feature_position)
            colors.append(feature_color)
        position_array = np.array(positions).flatten()
        color_array = np.array(colors).flatten()

        # create a tile
        feature_table_header = PntsFeatureTableHeader.from_semantic(
            SemanticPoint.POSITION, SemanticPoint.RGB, None, len(positions)
        )
        t = Pnts.from_features(feature_table_header, position_array, color_array)

        # configure the tile
        rtc = (1215012.8828876738, -4736313.051199594, 4081605.22126042)
        t.body.feature_table.header.rtc = rtc

        # get an array
        tile_arr = t.to_array()
        t2 = Pnts.from_array(tile_arr)
        self.assertEqual(t2.header.version, 1.0)

        # Test the tile byte length
        self.assertEqual(t2.header.tile_byte_length, 15176)
        self.assertEqual(
            t2.header.tile_byte_length % 8, 0
        )  # tile bytes must be 8-byte aligned

        # Test the feature table byte lengths
        json_feature_table_end = PntsHeader.BYTE_LENGTH + t2.header.ft_json_byte_length
        self.assertEqual(json_feature_table_end % 8, 0)
        self.assertEqual(t2.header.ft_json_byte_length, 148)
        bin_feature_table_end = json_feature_table_end + t2.header.ft_bin_byte_length
        self.assertEqual(bin_feature_table_end % 8, 0)
        self.assertEqual(t2.header.ft_bin_byte_length, 15000)

        self.assertEqual(t2.header.bt_json_byte_length, 0)
        self.assertEqual(t2.header.bt_bin_byte_length, 0)

        feature_table = t.body.feature_table
        pt_color = feature_table.get_feature_color_at(0)
        if pt_color is None:
            raise RuntimeError("pt_color should not be None")
        assert_array_equal(
            np.array([44, 243, 209], dtype=np.uint8),
            pt_color,
        )



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any

import numpy as np
import numpy.typing as npt

from py3dtiles.exceptions import TilerException
from py3dtiles.typing import BoundingVolumeBoxDictType

from .bounding_volume import BoundingVolume

if TYPE_CHECKING:
    from typing_extensions import Self

    from .tile import Tile

# In order to prevent the appearance of ghost newline characters ("\n")
# when printing a numpy.array (mainly self._box in this file):
np.set_printoptions(linewidth=500)


class BoundingVolumeBox(BoundingVolume[BoundingVolumeBoxDictType]):
    """
    A box bounding volume as defined in the 3DTiles specifications i.e. an
    array of 12 numbers that define an oriented bounding box:

    - The first three elements define the x, y, and z values for the center of the box.
    - The next three elements (with indices 3, 4, and 5) define the x axis
      direction and half-length.
    - The next three elements (with indices 6, 7, and 8) define the y axis
      direction and half-length.
    - The last three elements (indices 9, 10, and 11) define the z axis
      direction and half-length.

    Note that, by default, a box bounding volume doesn't need to be aligned
    with the coordinate axis. Still in general, computing the box bounding
    volume of two box bounding volumes won't necessarily yield a box that is
    aligned with the coordinate axis (although this computation might require
    some fitting algorithm e.g. the principal component analysis method.
    Yet in sake of simplification (and numerical efficiency), when asked to
    "add" (i.e. to find the enclosing box of) two (or more) box bounding
    volumes this class resolves to compute the "canonical" fitting/enclosing
    box i.e. a box that is parallel to the coordinate axis.
    """

    def __init__(self) -> None:
        super().__init__()
        self._box: npt.NDArray[np.float64] | None = None

    @classmethod
    def from_dict(cls, bounding_volume_box_dict: BoundingVolumeBoxDictType) -> Self:
        """
        Construct a BoundingVolumeBox from a dict following the structure of a 3dtiles bounding volume
        """
        bounding_volume_box = cls()
        bounding_volume_box.set_from_list(bounding_volume_box_dict["box"])

        bounding_volume_box.set_properties_from_dict(bounding_volume_box_dict)

        return bounding_volume_box

    @classmethod
    def from_points(cls, points: list[npt.NDArray[np.float64]]) -> BoundingVolumeBox:
        """
        Construct a bounding box enclosing all the points.

        Internally call `set_from_points`.

        :return: a new instance of BoundingVolumeBox
        """
        result = cls()
        result.set_from_points(points)
        return result

    @classmethod
    def from_list(cls, box_list: npt.ArrayLike) -> BoundingVolumeBox:
        """
        Construct a bounding box enclosing all the points.

        Internally call `set_from_points`.

        :return: a new instance of BoundingVolumeBox
        """
        result = cls()
        result.set_from_list(box_list)
        return result

    def get_center(self) -> npt.NDArray[np.float64]:
        if self._box is None:
            raise AttributeError("Bounding Volume Box is not defined.")

        return self._box[0:3]

    def translate(self, offset: npt.NDArray[np.float64]) -> None:
        """
        Translate the box center with the given offset "vector"

        :param offset: the 3D vector by which the box should be translated
        """
        if self._box is None:
            raise AttributeError("Bounding Volume Box is not defined.")

        self._box[:3] += offset[:3]

    def transform(self, transform: npt.NDArray[np.float64]) -> None:
        """
        Apply the provided transformation matrix (4x4) to the box

        :param transform: transformation matrix (4x4) to be applied
        """
        if self._box is None:
            raise AttributeError("Bounding Volume Box is not defined.")

        assert transform.shape == (4, 4)

        # FIXME: the following code only uses the first three coordinates
        # of the transformation matrix (and basically ignores the fourth
        # column of transform). This looks like some kind of mistake...
        rotation = np.array([transform[0][0:3], transform[1][0:3], transform[2][0:3]])

        center = self._box[0:3]
        x_half_axis = self._box[3:6]
        y_half_axis = self._box[6:9]
        z_half_axis = self._box[9:12]

        # Apply the rotation part to each element
        new_center = rotation.dot(center)
        new_x_half_axis = rotation.dot(x_half_axis)
        new_y_half_axis = rotation.dot(y_half_axis)
        new_z_half_axis = rotation.dot(z_half_axis)
        self._box = np.concatenate(
            (new_center, new_x_half_axis, new_y_half_axis, new_z_half_axis)
        )
        offset = transform[:, 3][0:3]
        self.translate(offset)

    def set_from_list(self, box_list: npt.ArrayLike) -> None:
        """
        Set the box from a list of coordinates closely matching the 3Dtiles spec.

        :param box_list: An array of 12 numbers that define an oriented bounding box. The first three elements define the x, y, and z values for the center of the box. The next three elements (with indices 3, 4, and 5) define the x axis direction and half-length. The next three elements (indices 6, 7, and 8) define the y axis direction and half-length. The last three elements (indices 9, 10, and 11) define the z axis direction and half-length.

        """
        box = np.array(box_list, dtype=np.float64)

        valid, reason = BoundingVolumeBox.is_valid(box)
        if not valid:
            raise ValueError(reason)
        self._box = box

    def set_from_points(self, points: list[npt.NDArray[np.float64]]) -> None:
        """
        Make the current box only include a list of points. Note: the box limits are replaced, not extended.

        :param points: An array of points
        """
        box = BoundingVolumeBox.get_box_array_from_point(points)

        valid, reason = BoundingVolumeBox.is_valid(box)
        if not valid:
            raise ValueError(reason)
        self._box = box

    def set_from_mins_maxs(self, mins_maxs: npt.NDArray[np.float64]) -> None:
        """
        Set the box from a min and a max.

        :param mins_maxs: the array [x_min, y_min, z_min, x_max, y_max, z_max]
                          that is the boundaries of the box along each
                          coordinate axis
        """
        self._box = BoundingVolumeBox.get_box_array_from_mins_maxs(mins_maxs)

    def get_corners(self) -> list[npt.NDArray[np.float64]]:
        """
        :return: the corners (3D points) of the box as a list
        """
        if self._box is None:
            raise AttributeError("Bounding Volume Box is not defined.")

        center, x_half_axis, y_half_axis, z_half_axis = self._box.reshape([-1, 3])

        x_axis = x_half_axis * 2
        y_axis = y_half_axis * 2
        z_axis = z_half_axis * 2

        # The eight cornering points of the box
        origin = center - x_half_axis - y_half_axis - z_half_axis

        ox = origin + x_axis
        oy = origin + y_axis
        oz = origin + z_axis
        oxy = ox + y_axis
        oxz = ox + z_axis
        oyz = oy + z_axis
        oxyz = oxy + z_axis

        return [origin, ox, oy, oxy, oz, oxz, oyz, oxyz]

    def get_canonical_as_array(self) -> npt.NDArray[np.float64]:
        """
        :return: the smallest enclosing box (as an array) that is parallel
                 to the coordinate axis
        """
        return BoundingVolumeBox.get_box_array_from_point(self.get_corners())

    def add(self, other: BoundingVolume[Any]) -> None:
        """
        Compute the 'canonical' bounding volume fitting this bounding volume
        together with the added bounding volume. Again (refer above to the
        class definition) the computed fitting bounding volume is generically
        not the smallest one (due to its alignment with the coordinate axis).

        :param other: another box bounding volume to be added with this one
        """
        if not isinstance(other, BoundingVolumeBox):
            raise NotImplementedError(
                "The add method works only with BoundingVolumeBox"
            )

        if self._box is None:
            # Then it is safe to overwrite
            self._box = copy.deepcopy(other._box)
            return

        corners = self.get_corners() + other.get_corners()
        self.set_from_points(corners)

    def sync_with_children(self, owner: Tile) -> None:
        # We reset to some dummy state of this Bounding Volume Box so we
        # can add up in place the boxes of the owner's children
        # If there is no child, no modifications are done.
        for child in owner.children:
            if child.bounding_volume is None:
                raise TilerException("Child should have a bounding volume.")

            bounding_volume = copy.deepcopy(child.bounding_volume)
            bounding_volume.transform(child.transform)
            if not isinstance(bounding_volume, BoundingVolumeBox):
                raise TilerException(
                    "All children must also have a box as bounding volume "
                    "if the parent has a bounding box"
                )
            self.add(bounding_volume)

    def to_dict(self) -> BoundingVolumeBoxDictType:
        if self._box is None:
            raise AttributeError("Bounding Volume Box is not defined.")

        dict_data: BoundingVolumeBoxDictType = {"box": list(self._box)}
        return self.add_root_properties_to_dict(dict_data)

    @staticmethod
    def get_box_array_from_mins_maxs(
        mins_maxs: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """
        :param mins_maxs: the list [x_min, y_min, z_min, x_max, y_max, z_max]
                          that is the boundaries of the box along each
                          coordinate axis
        :return: the smallest box (as an array, as opposed to a
                BoundingVolumeBox instance) that encloses the given list of
                (3D) points and that is parallel to the coordinate axis.
        """
        x_min = mins_maxs[0]
        x_max = mins_maxs[3]
        y_min = mins_maxs[1]
        y_max = mins_maxs[4]
        z_min = mins_maxs[2]
        z_max = mins_maxs[5]
        new_center = np.array(
            [(x_min + x_max) / 2, (y_min + y_max) / 2, (z_min + z_max) / 2]
        )
        new_x_half_axis = np.array([(x_max - x_min) / 2, 0, 0])
        new_y_half_axis = np.array([0, (y_max - y_min) / 2, 0])
        new_z_half_axis = np.array([0, 0, (z_max - z_min) / 2])

        return np.concatenate(
            (new_center, new_x_half_axis, new_y_half_axis, new_z_half_axis)
        )

    @staticmethod
    def get_box_array_from_point(
        points: list[npt.NDArray[np.float64]],
    ) -> npt.NDArray[np.float64]:
        """
        :param points: a list of 3D points
        :return: the smallest box (as an array, as opposed to a
                BoundingVolumeBox instance) that encloses the given list of
                (3D) points and that is parallel to the coordinate axis.
        """
        return BoundingVolumeBox.get_box_array_from_mins_maxs(
            np.array(
                [
                    min(c[0] for c in points),
                    min(c[1] for c in points),
                    min(c[2] for c in points),
                    max(c[0] for c in points),
                    max(c[1] for c in points),
                    max(c[2] for c in points),
                ]
            )
        )

    @staticmethod
    def is_valid(box: npt.NDArray[np.float64]) -> tuple[bool, str]:
        if box is None:
            return False, "Bounding Volume Box is not defined."
        if box.ndim != 1:
            return False, "Bounding Volume Box has wrong dimensions."
        if box.shape[0] != 12:
            return False, "Warning: Bounding Volume Box must have 12 elements."
        return True, ""
